#!/usr/bin/python3

import collections
import errno
import ledger
from ledgerhelpers import parser, debug_time
import ledgerhelpers.legacy_needsledger as hln
import logging
from multiprocessing import Process, Pipe
import os
import threading
import time


UNCHANGED = "unchanged"
UNCONDITIONAL = "unconditional"
IFCHANGED = "ifchanged"

CMD_GET_A_LCFA_C = "get_accounts_last_commodity_for_account_and_commodities"


def transactions_with_payee(payee,
                            internal_parsing_result,
                            case_sensitive=True):
    """Given a payee string, and an internal_parsing() result from the
    journal, return the transactions with matching payee,
    perhaps ignoring case."""
    transes = []
    if not case_sensitive:
        payee = payee.lower()
    for xact in internal_parsing_result:
        if not hasattr(xact, "payee"):
            continue
        xact_payee = xact.payee if case_sensitive else xact.payee.lower()
        if xact_payee == payee:
            transes.append(xact)
    return transes


class Joinable(threading.Thread):
    """A subclass of threading.Thread that catches exception in run(), if any,
    and re-throws it in join()."""
    exception = None

    def __init__(self):
        threading.Thread.__init__(self, daemon=True)

    def join(self):
        threading.Thread.join(self)
        if self.exception:
            raise self.exception

    def run(self):
        try:
            self.__run__()
        except BaseException as e:
            self.exception = e


class JournalCommon():

    path = None
    path_mtime = None
    price_path = None
    price_path_mtime = None

    def changed(self):
        path_mtime = None
        price_path_mtime = None

        try:
            path_mtime = os.stat(self.path).st_mtime
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        try:
            if self.price_path is not None:
                price_path_mtime = os.stat(self.price_path).st_mtime
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

        if (
            path_mtime != self.path_mtime or
            price_path_mtime != self.price_path_mtime
        ):
            self.path_mtime = path_mtime
            self.price_path_mtime = price_path_mtime
            self.logger.debug("Files have changed, rereading.")
            return True
        else:
            return False

    def _get_text(self, prepend_price_path=False):
        files = []
        if self.price_path and prepend_price_path:
            files.append(self.price_path)
        if self.path:
            files.append(self.path)
        t = []
        for f in files:
            with open(f, "r") as fo:
                t.append(fo.read())
        text = "\n".join(t)
        self.logger.debug("Read %d characters of journal%s.", len(text), " and price file" if len(files) > 1 else "")
        return text

    def get_journal_text_with_prices(self):
        return self._get_text(True)

    def get_journal_text(self):
        return self._get_text(False)


class Journal(JournalCommon):

    logger = logging.getLogger("journal.master")

    pipe = None
    slave = None
    slave_lock = None
    cache = None
    internal_parsing_cache = None
    internal_parsing_cache_lock = None

    def __init__(self):
        """Do not instantiate directly.  Use class methods."""
        self.cache = {}
        self.internal_parsing_cache = []
        self.internal_parsing_cache_lock = threading.Lock()
        self.slave_lock = threading.Lock()

    def _start_slave(self):
        if self.pipe:
            self.pipe.close()
        self.pipe, theirconn = Pipe()
        if self.slave:
            try:
                self.slave.terminate()
            except Exception:
                pass
        try:
            self.slave = JournalSlave(theirconn, self.path, self.price_path)
            self.slave.start()
        finally:
            theirconn.close()

    @classmethod
    def from_file(klass, journal_file, price_file):
        j = klass()
        j.path = journal_file
        j.price_path = price_file
        j._start_slave()
        j._cache_internal_parsing()
        return j

    def _cache_internal_parsing(self):
        self.internal_parsing_cache_lock.acquire()

        if self.changed():
            me = self
            self.internal_parsing_cache = None

            class Rpi(Joinable):
                @debug_time(self.logger)
                def __run__(self):
                    try:
                        me.logger.debug("Reparsing internal.")
                        res = parser.lex_ledger_file_contents(me.get_journal_text())
                        me.internal_parsing_cache = res
                    finally:
                        me.internal_parsing_cache_lock.release()

            internal_parsing_thread = Rpi()
            internal_parsing_thread.name = "Internal reparser"
            internal_parsing_thread.start()
            return internal_parsing_thread
        else:
            # Dummy thread.  Just serves to unlock the parsing cache lock.
            nothread = threading.Thread(target=self.internal_parsing_cache_lock.release)
            nothread.start()
            return nothread

    def _cache_accounts_last_commodity_for_account_and_commodities(self):
        with self.slave_lock:
            try:
                self.pipe.send(
                    (CMD_GET_A_LCFA_C,
                     IFCHANGED if "accounts" in self.cache
                     else UNCONDITIONAL)
                )
                result = self.pipe.recv()
                if isinstance(result, BaseException):
                    raise result
                if result == UNCHANGED:
                    assert "accounts" in self.cache
                else:
                    accounts = result[0] if result[0] is not None else []
                    last_commodity_for_account = dict(
                        (acc, ledger.Amount(amt))
                        for acc, amt in list(result[1].items())
                    ) if result[1] is not None else {}
                    all_commodities = [
                        ledger.Amount(c)
                        for c in result[2]
                    ] if result[2] is not None else []
                    self.cache["accounts"] = accounts
                    self.cache["last_commodity_for_account"] = (
                        last_commodity_for_account
                    )
                    self.cache["all_commodities"] = all_commodities
            except BaseException:
                self.cache = {}
                self._start_slave()
                raise

    @debug_time(logger)
    def accounts_and_last_commodity_for_account(self):
        self._cache_accounts_last_commodity_for_account_and_commodities()
        return self.cache["accounts"], self.cache["last_commodity_for_account"]

    @debug_time(logger)
    def commodities(self):
        self._cache_accounts_last_commodity_for_account_and_commodities()
        return self.cache["all_commodities"]

    def commodity(self, label, create=False):
        pool = ledger.Amount("$ 1").commodity.pool()
        if create:
            return pool.find_or_create(label)
        else:
            return pool.find(label)

    @debug_time(logger)
    def all_payees(self):
        """Returns a list of strings with payees (transaction titles)."""
        titles = collections.OrderedDict()
        for xact in self.internal_parsing():
            if hasattr(xact, "payee") and xact.payee not in titles:
                titles[xact.payee] = xact.payee
        return list(titles.keys())

    @debug_time(logger)
    def internal_parsing(self):
        self._cache_internal_parsing().join()
        with self.internal_parsing_cache_lock:
            return self.internal_parsing_cache

    def generate_record(self, *args):
        return hln.generate_record(*args)

    def generate_price_records(self, prices):
        from ledgerhelpers import generate_price_records
        return generate_price_records(prices)

    def _add_text_to_file(self, text, f):
        if not isinstance(text, str):
            text = "\n".join(text)
        f = open(f, "a")
        print(text, file=f)
        f.flush()
        f.close()

    def add_text_to_file(self, text):
        return self._add_text_to_file(text, self.path)

    def add_text_to_price_file(self, text):
        return self._add_text_to_file(text, self.price_path)


class JournalSlave(JournalCommon, Process):

    session = None
    journal = None
    accounts = None
    last_commodity_for_account = None
    all_commodities = None
    logger = logging.getLogger("journal.slave")

    def __init__(self, pipe, path, price_path):
        Process.__init__(self)
        self.daemon = True
        self.pipe = pipe
        self.path = path
        self.price_path = price_path
        self.clear_caches()

    def clear_caches(self):
        self.session = None
        self.journal = None
        self.accounts = None
        self.last_commodity_for_account = None
        self.all_commodities = None

    def reparse_ledger(self):
        self.logger.debug("Reparsing ledger.")
        session = ledger.Session()
        journal = session.read_journal_from_string(self.get_journal_text_with_prices())
        self.session = session
        self.journal = journal

    def harvest_accounts_and_last_commodities(self):
        self.logger.debug("Harvesting accounts and last commodities.")
        # Commodities returned by this method do not contain any annotations.
        accts = []
        commos = dict()
        amts = dict()
        for post in self.journal.query(""):
            for subpost in post.xact.posts():
                if str(subpost.account) not in accts:
                    accts.append(str(subpost.account))
                comm = subpost.amount / subpost.amount
                comm.commodity = comm.commodity.strip_annotations()
                commos[str(subpost.account)] = str(comm)
                amts[str(comm)] = True
        self.accounts = accts
        self.last_commodity_for_account = commos
        self.all_commodities = [str(k) for k in list(amts.keys())]

    def reparse_all_if_needed(self):
        me = self

        changed = self.changed()
        if changed:
            self.clear_caches()

            class Rpl(Joinable):
                @debug_time(self.logger)
                def __run__(self):
                    me.reparse_ledger()
                    me.harvest_accounts_and_last_commodities()

            ledger_parsing_thread = Rpl()
            ledger_parsing_thread.name = "Ledger reparser"

        else:
            ledger_parsing_thread = threading.Thread(target=len, args=([],))
            ledger_parsing_thread.name = "Dummy ledger reparser"

        ledger_parsing_thread.start()

        return (
            changed, ledger_parsing_thread
        )

    def run(self):
        logger = logging.getLogger("journal.slave.loop")
        _, initial_parsing_thread = self.reparse_all_if_needed()
        while True:
            cmd_args = self.pipe.recv()
            if initial_parsing_thread:
                initial_parsing_thread.join()
                initial_parsing_thread = None
            cmd = cmd_args[0]
            start = time.time()
            logger.debug("* Servicing: %-55s  started", cmd)
            args = cmd_args[1:]
            try:
                changed, lpt = self.reparse_all_if_needed()
                if cmd == CMD_GET_A_LCFA_C:
                    if (
                        not changed and
                        args[0] == IFCHANGED and
                        self.journal
                    ):
                        logger.debug("* Serviced:  %-55s  %.3f seconds - %s",
                                     cmd, time.time() - start, UNCHANGED)
                        self.pipe.send(UNCHANGED)
                        continue
                    lpt.join()
                    logger.debug("* Serviced:  %-55s  %.3f seconds - new data",
                                 cmd, time.time() - start)
                    self.pipe.send((
                        self.accounts,
                        self.last_commodity_for_account,
                        self.all_commodities,
                    ))
                else:
                    assert 0, "not reached"
            except BaseException as e:
                logger.exception("Unrecoverable error in slave.")
                self.pipe.send(e)



#!/usr/bin/python3

# This code is imported straight from the Kiwi codebase, and ported to work
# with GTK+ 3.x.
#
# The original source code is here:
# http://doc.stoq.com.br/api/kiwi/_modules/kiwi/ui/dateentry.html
#
# The original program was provided under the LGPL 2.1 or later license terms.
# As such, this program reuses it freely, linking it with the program is
# permitted and should be no issue.

import datetime

from gi.repository import GObject
import gi
gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Gdk
from gi.repository import Gtk

from ledgerhelpers.legacy import format_date, parse_date


def prev_month(date):
    if date.month == 1:
        return datetime.date(date.year - 1, 12, date.day)
    else:
        day = date.day
        while True:
            try:
                return datetime.date(date.year, date.month - 1, day)
            except ValueError:
                day = day - 1


def next_month(date):
    if date.month == 12:
        return datetime.date(date.year + 1, 1, date.day)
    else:
        day = date.day
        while True:
            try:
                return datetime.date(date.year, date.month + 1, day)
            except ValueError:
                day = day - 1


def beginning_of_month(date):
    return datetime.date(date.year, date.month, 1)


def end_of_month(date):
    if date.month == 12:
        date = datetime.date(date.year + 1, 1, 1)
    else:
        date = datetime.date(date.year, date.month + 1, 1)
    return date - datetime.timedelta(1)


def prev_day(date):
    return date - datetime.timedelta(1)


def next_day(date):
    return date + datetime.timedelta(1)


def _according_to_keyval(keyval, state, date, skip="", in_editbox=False):
    if (keyval in (Gdk.KEY_Page_Up, Gdk.KEY_KP_Page_Up)):
        if date:
            return True, prev_month(date)
        else:
            return True, None
    if (keyval in (Gdk.KEY_Page_Down, Gdk.KEY_KP_Page_Down)):
        if date:
            return True, next_month(date)
        else:
            return True, None
    if (
        keyval in (Gdk.KEY_minus, Gdk.KEY_KP_Subtract, Gdk.KEY_underscore) and
        "minus" not in skip
    ):
        if date:
            return True, prev_day(date)
        else:
            return True, None
    if (keyval in (Gdk.KEY_plus, Gdk.KEY_KP_Add,
                   Gdk.KEY_equal, Gdk.KEY_KP_Equal)):
        if date:
            return True, next_day(date)
        else:
            return True, None
    if (
        keyval in (Gdk.KEY_Home, Gdk.KEY_KP_Home) and
        (not in_editbox or state & Gdk.ModifierType.SHIFT_MASK)
    ):
        if date:
            return True, beginning_of_month(date)
        else:
            return True, None
    if (
        keyval in (Gdk.KEY_End, Gdk.KEY_KP_End) and
        (not in_editbox or state & Gdk.ModifierType.SHIFT_MASK)
    ):
        if date:
            return True, end_of_month(date)
        else:
            return True, None
    return False, None


class _DateEntryPopup(Gtk.Window):

    __gsignals__ = {
        'date-selected': (
            GObject.SIGNAL_RUN_LAST,
            None,
            (object,)
        )
    }

    def __init__(self, dateentry):
        Gtk.Window.__init__(self, Gtk.WindowType.POPUP)
        self.add_events(Gdk.EventMask.BUTTON_PRESS_MASK)
        self.connect('key-press-event', self._on__key_press_event)
        self.connect('button-press-event', self._on__button_press_event)
        self._dateentry = dateentry

        frame = Gtk.Frame()
        frame.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        self.add(frame)
        frame.show()

        vbox = Gtk.VBox()
        vbox.set_border_width(12)
        frame.add(vbox)
        vbox.show()
        self._vbox = vbox

        self.calendar = Gtk.Calendar()
        self.calendar.connect('day-selected-double-click',
                              self._on_calendar__day_selected_double_click)
        self.calendar.connect('day-selected',
                              self._on_calendar__day_selected)
        vbox.pack_start(self.calendar, False, False, 0)
        self.calendar.show()

        buttonbox = Gtk.HButtonBox()
        buttonbox.set_border_width(12)
        buttonbox.set_layout(Gtk.ButtonBoxStyle.SPREAD)
        vbox.pack_start(buttonbox, False, False, 0)
        buttonbox.show()

        for label, callback in [('_Today', self._on_today__clicked),
                                ('_Close', self._on_close__clicked)]:
            button = Gtk.Button(label, use_underline=True)
            button.connect('clicked', callback)
            buttonbox.pack_start(button, False, False, 0)
            button.show()

        self.set_resizable(False)
        self.set_screen(dateentry.get_screen())

        self.realize()
        self.height = self._vbox.size_request().height

    def _on_calendar__day_selected_double_click(self, unused_calendar):
        self.emit('date-selected', self.get_date())
        self.popdown()

    def _on_calendar__day_selected(self, unused_calendar):
        self.emit('date-selected', self.get_date())

    def _on__button_press_event(self, unused_window, event):
        # If we're clicking outside of the window close the popup
        hide = False

        # Also if the intersection of self and the event is empty, hide
        # the calendar
        if (tuple(self.get_allocation().intersect(
            Gdk.Rectangle(x=int(event.x), y=int(event.y),
                          width=1, height=1))) == (0, 0, 0, 0)):
            hide = True

        # Toplevel is the window that received the event, and parent is the
        # calendar window. If they are not the same, means the popup should
        # be hidden. This is necessary for when the event happens on another
        # widget
        toplevel = event.get_window().get_toplevel()
        parent = self.calendar.get_parent_window()
        if toplevel != parent:
            hide = True

        if hide:
            self.popdown()

    def _on__key_press_event(self, unused_window, event):
        """
        Mimics Combobox behavior

        Escape, Enter or Alt+Up: Close
        Space: Select
        """
        keyval = event.keyval
        state = event.state & Gtk.accelerator_get_default_mod_mask()
        if (keyval == Gdk.KEY_Escape or
            keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter) or
            ((keyval == Gdk.KEY_Up or keyval == Gdk.KEY_KP_Up) and
             state == Gdk.ModifierType.MOD1_MASK)):
            self.popdown()
            return True
        processed, new_date = _according_to_keyval(keyval, state,
                                                   self.get_date())
        if processed and new_date:
            self.set_date(new_date)
        return processed

    def _on_select__clicked(self, unused_button):
        self.emit('date-selected', self.get_date())

    def _on_close__clicked(self, unused_button):
        self.popdown()

    def _on_today__clicked(self, unused_button):
        self.set_date(datetime.date.today())

    def _popup_grab_window(self):
        activate_time = 0
        win = self.get_window()
        result = Gdk.pointer_grab(win, True, (
            Gdk.EventMask.BUTTON_PRESS_MASK |
            Gdk.EventMask.BUTTON_RELEASE_MASK |
            Gdk.EventMask.POINTER_MOTION_MASK),
            None, None, activate_time
        )
        if result == 0:
            if Gdk.keyboard_grab(self.get_window(), True, activate_time) == 0:
                return True
            else:
                self.get_window().get_display().pointer_ungrab(activate_time)
                return False
        return False

    def _get_position(self):
        self.realize()
        calendar = self

        sample = self._dateentry

        # We need to fetch the coordinates of the entry window
        # since comboentry itself does not have a window
        origin = sample.entry.get_window().get_origin()
        x, y = origin.x, origin.y
        width = calendar.size_request().width
        height = self.height

        screen = sample.get_screen()
        monitor_num = screen.get_monitor_at_window(sample.get_window())
        monitor = screen.get_monitor_geometry(monitor_num)

        if x < monitor.x:
            x = monitor.x
        elif x + width > monitor.x + monitor.width:
            x = monitor.x + monitor.width - width

        alloc = sample.get_allocation()

        if y + alloc.height + height <= monitor.y + monitor.height:
            y += alloc.height
        elif y - height >= monitor.y:
            y -= height
        elif (monitor.y + monitor.height - (y + alloc.height) >
              y - monitor.y):
            y += alloc.height
            height = monitor.y + monitor.height - y
        else:
            height = y - monitor.y
            y = monitor.y

        return x, y, width, height

    def popup(self, date):
        """
        Shows the list of options. And optionally selects an item
        :param date: date to select
        """
        combo = self._dateentry
        if not (combo.get_realized()):
            return

        treeview = self.calendar
        if treeview.get_mapped():
            return
        toplevel = combo.get_toplevel()
        if isinstance(toplevel, Gtk.Window) and toplevel.get_group():
            toplevel.get_group().add_window(self)

        x, y, width, height = self._get_position()
        self.set_size_request(width, height)
        self.move(x, y)

        if date is not None:
            self.set_date(date)
        self.grab_focus()

        if not (self.calendar.has_focus()):
            self.calendar.grab_focus()

        self.show_all()
        if not self._popup_grab_window():
            self.hide()
            return

        self.grab_add()

    def popdown(self):
        """Hides the list of options"""
        combo = self._dateentry
        if not (combo.get_realized()):
            return

        self.grab_remove()
        self.hide()

    # month in gtk.Calendar is zero-based (i.e the allowed values are 0-11)
    # datetime one-based (i.e. the allowed values are 1-12)
    # So convert between them

    def get_date(self):
        """Gets the date of the date entry
        :returns: date of the entry
        :rtype date: datetime.date
        """
        y, m, d = self.calendar.get_date()
        return datetime.date(y, m + 1, d)

    def set_date(self, date):
        """Sets the date of the date entry
        :param date: date to set
        :type date: datetime.date
        """
        self.calendar.select_month(date.month - 1, date.year)
        self.calendar.select_day(date.day)
        # FIXME: Only mark the day in the current month?
        self.calendar.clear_marks()
        self.calendar.mark_day(date.day)


@GObject.type_register
class DateEntry(Gtk.HBox):
    """I am an entry which you can input a date on.
    I make entering a date blazing fast.

    The date you input in me must be of the form YYYY-MM-DD or YYYY/MM/DD.
    These are the date formats expected by Ledger.

    In addition to the text box where you can type, I also contain a button
    with an icon you can click, to get a popup window with a date picker
    where you can select the date.  You can also show the date picker
    by focusing me and then hitting Alt+Down on your keyboard.

    There are a number of cool combos you can use, whether in the text box
    or in the date picker popup:

    * Minus:\t\tprevious day
    * Plus:\t\tnext day
    * Page Up:\tprevious month
    * Page Down:\tnext month
    * Home:\t\tbeginning of the month  (if textbox is focused, Shift+Home)
    * End:\t\tend of the month (if textbox is focused, Shift+End)

    In the date picker popup, hitting Enter, or Escape, or Alt+Up after
    selecting a date (making it blue with a click or with the Space bar)
    makes the date picker popup go away.  Clicking outside the date picker
    popup also closes it.
    """

    __gsignals__ = {
        'changed': (
            GObject.SIGNAL_RUN_LAST,
            None,
            ()
        ),
        'activate': (
            GObject.SIGNAL_RUN_LAST,
            None,
            ()
        ),
    }

    def __init__(self):
        Gtk.HBox.__init__(self)

        self._old_date = None

        self.entry = Gtk.Entry()
        self.entry.set_max_length(10)
        self.entry.set_width_chars(13)
        self.entry.set_overwrite_mode(True)
        self.entry.set_tooltip_text(
            self.__doc__.replace("\n    ", "\n").strip()
        )
        self.entry.set_property('primary-icon-name', "x-office-calendar")
        self.entry.connect('changed', self._on_entry__changed)
        self.entry.connect('activate', self._on_entry__activate)
        self.entry.connect('key-press-event', self._on_entry__key_press_event)
        self.entry.connect('icon-press', self._on_entry__icon_press)
#       ADD SUPPORT
#       self._button.connect('scroll-event', self._on_entry__scroll_event)
        self.entry.set_placeholder_text("Date")
        self.pack_start(self.entry, False, False, 0)
        self.entry.show()

        self._popup = _DateEntryPopup(self)
        self._popup.connect('date-selected', self._on_popup__date_selected)
        self._popup.connect('hide', self._on_popup__hide)
        self._popup.set_size_request(-1, 24)

    def _on_entry__icon_press(self, unused_entry,
                              entry_icon_position,
                              event_button):
        if (
            entry_icon_position == Gtk.EntryIconPosition.PRIMARY and
            event_button.get_button() == (True, 1)
        ):
            self._popup_date_picker()
            return True

    def _on_entry__key_press_event(self, unused_window, event):
        keyval = event.keyval
        state = event.state & Gtk.accelerator_get_default_mod_mask()
        if (
            (keyval == Gdk.KEY_Down or keyval == Gdk.KEY_KP_Down) and
            state == Gdk.ModifierType.MOD1_MASK
        ):
            self._popup_date_picker()
            return True

        skip_minus = ""
        if not self.get_date():
            skip_minus = "minus"
        if self.entry.get_property("cursor-position") in (4, 7):
            skip_minus = "minus"
        processed, new_date = _according_to_keyval(keyval,
                                                   state,
                                                   self.get_date(),
                                                   skip=skip_minus,
                                                   in_editbox=True)
        if processed and new_date:
            self.set_date(new_date)
        return processed

    # Virtual methods

    def do_grab_focus(self):
        self.entry.grab_focus()

    # Callbacks

    def _on_entry__changed(self, unused_entry):
        try:
            date = self.get_date()
        except ValueError:
            date = None
        self._changed(date)

    def _on_entry__activate(self, unused_entry):
        self.emit('activate')

    def _on_entry__scroll_event(self, unused_entry, event):
        if event.direction == Gdk.SCROLL_UP:
            days = 1
        elif event.direction == Gdk.SCROLL_DOWN:
            days = -1
        else:
            return

        try:
            date = self.get_date()
        except ValueError:
            date = None

        if not date:
            newdate = datetime.date.today()
        else:
            newdate = date + datetime.timedelta(days=days)
        self.set_date(newdate)

    def _popup_date_picker(self):
        try:
            date = self.get_date()
        except ValueError:
            date = None
        self._popup.popup(date)

    def _on_popup__hide(self, popup):
        pass

    def _on_popup__date_selected(self, unused_popup, date):
        self.set_date(date)
        self.entry.grab_focus()
        self.entry.set_position(len(self.entry.get_text()))
        self._changed(date)

    def _changed(self, date):
        if self._old_date != date:
            self.emit('changed')
            self._old_date = date

    # Public API

    def set_activates_default(self, val):
        return self.entry.set_activates_default(val)

    def set_date(self, date):
        """Sets the date.
        :param date: date to set
        :type date: a datetime.date instance or None
        """
        if not isinstance(date, datetime.date) and date is not None:
            raise TypeError(
                "date must be a datetime.date instance or None, not %r" % (
                    date,
                )
            )

        if date is None:
            value = ''
        else:
            try:
                value = format_date(date, self.entry.get_text())
            except ValueError:
                value = format_date(date, "2016-12-01")
        self.entry.set_text(value)

    def get_date(self):
        """Get the selected date
        :returns: the date.
        :rtype: datetime.date or None
        """
        try:
            date = self.entry.get_text()
            date = parse_date(date)
        except ValueError:
            date = None
        return date

    def follow(self, other_calendar):
        self.followed = other_calendar
        self.followed_last_value = other_calendar.get_date()

        def copy_when(other_calendar, *unused_args):
            if (
                self.get_date() == self.followed_last_value or
                other_calendar.get_date() > self.get_date()
            ):
                self.set_date(other_calendar.get_date())
            self.followed_last_value = other_calendar.get_date()
        other_calendar.connect("changed", copy_when)


class TestWindow(Gtk.Window):

    def __init__(self):
        Gtk.Window.__init__(self, title="Whatever")
        self.set_border_width(12)

        combobox = DateEntry()
        self.add(combobox)


def main():
    klass = TestWindow
    win = klass()
    win.connect("delete-event", Gtk.main_quit)
    GObject.idle_add(win.show_all)
    Gtk.main()


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import ledgerhelpers.legacy
from ledgerhelpers import diffing


CHAR_ENTER = "\n"
CHAR_COMMENT = ";#"
CHAR_NUMBER = "1234567890"
CHAR_TAB = "\t"
CHAR_WHITESPACE = " \t"
CHAR_CLEARED = "*"
CHAR_PENDING = "!"

STATE_CLEARED = CHAR_CLEARED
STATE_PENDING = CHAR_PENDING
STATE_UNCLEARED = None


def pos_within_items_to_row_and_col(pos, items):
    row = 1
    col = 1
    for i, c in enumerate(items):
        if i >= pos:
            break
        if c in CHAR_ENTER:
            row += 1
            col = 1
        else:
            col += 1
    return row, col


def parse_date_from_transaction_contents(contents):
    return ledgerhelpers.legacy.parse_date("".join(contents))


class Token(object):

    def __init__(self, pos, contents):
        self.pos = pos
        if not isinstance(contents, str):
            contents = "".join(contents)
        self.contents = contents

    def __str__(self):
        return """<%s at pos %d len %d
%s>""" % (self.__class__.__name__, self.pos, len(self.contents), self.contents)


class TokenComment(Token):
    pass


class TokenTransactionComment(Token):
    pass


class TokenTransactionClearedFlag(Token):
    pass


class TokenTransactionPendingFlag(Token):
    pass


class TokenWhitespace(Token):
    pass


class TokenTransaction(Token):

    def __init__(self, pos, contents):
        Token.__init__(self, pos, contents)
        lexer = LedgerTransactionLexer(contents)
        lexer.run()

        def find_token(klass):
            try:
                return [t for t in lexer.tokens if isinstance(t, klass)][0]
            except IndexError:
                return None

        try:
            self.date = find_token(TokenTransactionDate).date
        except AttributeError:
            raise TransactionLexingError("no transaction date in transaction")

        try:
            self.secondary_date = find_token(
                TokenTransactionSecondaryDate
            ).date
        except AttributeError:
            self.secondary_date = None

        if find_token(TokenTransactionClearedFlag):
            self.state = STATE_CLEARED
        elif find_token(TokenTransactionPendingFlag):
            self.state = STATE_PENDING
        else:
            self.state = STATE_UNCLEARED

        if self.state != STATE_UNCLEARED:
            self.clearing_date = (
                self.secondary_date if self.secondary_date else self.date
            )
        else:
            self.clearing_date = None

        try:
            self.payee = find_token(TokenTransactionPayee).payee
        except AttributeError:
            raise TransactionLexingError("no payee in transaction")

        accountsamounts = [
            t for t in lexer.tokens
            if isinstance(t, TokenTransactionPostingAccount) or
            isinstance(t, TokenTransactionPostingAmount)
        ]

        x = []
        last = None
        for v in accountsamounts:
            if isinstance(v, TokenTransactionPostingAccount):
                assert type(last) in [
                     type(None), TokenTransactionPostingAmount
                ], lexer.tokens
            elif isinstance(v, TokenTransactionPostingAmount):
                assert type(last) in [
                    TokenTransactionPostingAccount
                ], lexer.tokens
                x.append(
                    ledgerhelpers.TransactionPosting(
                        last.account, v.amount
                    )
                )
            last = v
        assert len(x) * 2 == len(accountsamounts), lexer.tokens
        self.postings = x


class TokenTransactionWithContext(TokenTransaction):

    def __init__(self, pos, tokens):
        self.transaction = [
            t for t in tokens if isinstance(t, TokenTransaction)
        ][0]
        self.pos = pos
        self.contents = "".join(t.contents for t in tokens)

    @property
    def date(self):
        return self.transaction.date


class TokenConversion(Token):
    pass


class TokenPrice(Token):
    pass


class TokenEmbeddedPython(Token):
    pass


class TokenTransactionPostingAccount(Token):

    def __init__(self, pos, contents):
        Token.__init__(self, pos, contents)
        self.account = ''.join(contents)


class TokenTransactionPostingAmount(Token):

    def __init__(self, pos, contents):
        Token.__init__(self, pos, contents)
        self.amount = ''.join(contents)


class TokenEmbeddedTag(Token):
    pass


class TokenTransactionDate(Token):

    def __init__(self, pos, contents):
        Token.__init__(self, pos, contents)
        self.date = parse_date_from_transaction_contents(self.contents)


class TokenTransactionSecondaryDate(Token):

    def __init__(self, pos, contents):
        Token.__init__(self, pos, contents)
        self.date = parse_date_from_transaction_contents(self.contents)


class TokenTransactionPayee(Token):

    def __init__(self, pos, contents):
        Token.__init__(self, pos, contents)
        self.payee = ''.join(contents)


class LexingError(Exception):
    pass


class TransactionLexingError(Exception):
    pass


class EOF(LexingError):
    pass


class GenericLexer(object):

    def __init__(self, items):
        if isinstance(items, str) and not isinstance(items, str):
            self.items = tuple(items.decode("utf-8"))
        else:
            self.items = tuple(items)
        self.start = 0
        self.pos = 0
        self._last_emitted_pos = self.pos
        self.tokens = []

    def __next__(self):
        """Returns the item at the current position, and advances the position."""
        try:
            t = self.items[self.pos]
        except IndexError:
            raise EOF()
        self.pos += 1
        return t

    def peek(self):
        """Returns the item at the current position."""
        try:
            t = self.items[self.pos]
        except IndexError:
            raise EOF()
        return t

    def confirm_next(self, seq):
        """Returns True if each item in seq matches each corresponding item
        from the current position onward."""
        for n, i in enumerate(seq):
            try:
                if self.items[self.pos + n] != i:
                    return False
            except IndexError:
                return False
        return True

    def emit(self, klass, items):
        """Creates an instance of klass (a Token class) with the current
        position and the supplied items as parameters, then
        accumulates the instance into the self.tokens accumulator."""
        token = klass(self.pos, items)
        self._last_emitted_pos = self.pos
        self.tokens += [token]

    def more(self):
        return self.pos < len(self.items)


class LedgerTextLexer(GenericLexer):

    def __init__(self, text):
        assert isinstance(text, str), type(text)
        GenericLexer.__init__(self, text)

    def state_parsing_toplevel_text(self):
        """Returns another state function."""
        chars = []
        while self.more():
            if self.peek() in CHAR_COMMENT:
                self.emit(TokenWhitespace, chars)
                return self.state_parsing_comment
            if self.peek() in CHAR_NUMBER:
                self.emit(TokenWhitespace, chars)
                return self.state_parsing_transaction
            if self.confirm_next("P"):
                self.emit(TokenWhitespace, chars)
                return self.state_parsing_price
            if self.confirm_next("C"):
                self.emit(TokenWhitespace, chars)
                return self.state_parsing_conversion
            if self.confirm_next("python"):
                self.emit(TokenWhitespace, chars)
                return self.state_parsing_embedded_python
            if self.confirm_next("tag"):
                self.emit(TokenWhitespace, chars)
                return self.state_parsing_embedded_tag
            if self.peek() not in CHAR_WHITESPACE + CHAR_ENTER:
                _, _, l2, c2 = self._coords()
                raise LexingError(
                    "unparsable data at line %d, char %d" % (l2, c2)
                )
            chars += [next(self)]
        self.emit(TokenWhitespace, chars)
        return

    def state_parsing_comment(self):
        chars = [next(self)]
        while self.more():
            if chars[-1] in CHAR_ENTER and self.peek() not in CHAR_COMMENT:
                break
            chars.append(next(self))
        self.emit(TokenComment, chars)
        return self.state_parsing_toplevel_text

    def state_parsing_price(self):
        return self.state_parsing_embedded_directive(TokenPrice, False)

    def state_parsing_conversion(self):
        return self.state_parsing_embedded_directive(TokenConversion, False)

    def state_parsing_embedded_tag(self):
        return self.state_parsing_embedded_directive(TokenEmbeddedTag)

    def state_parsing_embedded_python(self):
        return self.state_parsing_embedded_directive(TokenEmbeddedPython)

    def state_parsing_embedded_directive(self, klass, maybe_multiline=True):
        chars = [next(self)]
        while self.more():
            if chars[-1] in CHAR_ENTER:
                if not maybe_multiline:
                    break
                if self.peek() in CHAR_WHITESPACE + CHAR_ENTER:
                    chars.append(next(self))
                    continue
                if self.peek() in CHAR_COMMENT:
                    self.emit(klass, chars)
                    return self.state_parsing_comment
                if self.peek() in CHAR_NUMBER:
                    self.emit(klass, chars)
                    return self.state_parsing_transaction
                self.emit(klass, chars)
                return self.state_parsing_toplevel_text
            chars.append(next(self))
        self.emit(klass, chars)
        return self.state_parsing_toplevel_text

    def state_parsing_transaction(self):
        chars = [next(self)]
        while self.more():
            if chars[-1] in CHAR_ENTER and self.peek() not in CHAR_WHITESPACE:
                break
            chars.append(next(self))
        self.emit(TokenTransaction, chars)
        return self.state_parsing_toplevel_text

    def _coords(self):
        r, c = pos_within_items_to_row_and_col(self._last_emitted_pos, self.items)
        r2, c2 = pos_within_items_to_row_and_col(self.pos, self.items)
        return r, c, r2, c2

    def run(self):
        state = self.state_parsing_toplevel_text
        while state:
            try:
                state = state()
            except LexingError:
                raise
            except Exception as e:
                l, c, l2, c2 = self._coords()
                raise LexingError(
                    "bad ledger data between line %d, char %d and line %d, char %d: %s" % (
                        l, c, l2, c2, e
                    )
                )


class LedgerTransactionLexer(GenericLexer):

    def __init__(self, text):
        GenericLexer.__init__(self, text)

    def state_parsing_transaction_date(self):
        chars = []
        while self.more():
            if self.peek() not in "0123456789-/":
                self.emit(TokenTransactionDate, chars)
                if self.confirm_next("="):
                    next(self)
                    return self.state_parsing_clearing_date
                elif self.peek() in CHAR_WHITESPACE:
                    return self.state_parsing_cleared_flag_or_payee
                else:
                    raise TransactionLexingError("invalid character %s" % self.peek())
            chars += [next(self)]
        raise TransactionLexingError("incomplete transaction")

    def state_parsing_clearing_date(self):
        chars = []
        while self.more():
            if self.peek() not in "0123456789-/":
                next(self)
                self.emit(TokenTransactionSecondaryDate, chars)
                return self.state_parsing_cleared_flag_or_payee
            chars += [next(self)]
        raise TransactionLexingError("incomplete transaction")

    def state_parsing_cleared_flag_or_payee(self):
        while self.more():
            if self.peek() in CHAR_WHITESPACE:
                next(self)
                continue
            if self.peek() in CHAR_ENTER:
                break
            if self.confirm_next(CHAR_CLEARED):
                self.emit(TokenTransactionClearedFlag, [next(self)])
                return self.state_parsing_payee
            if self.confirm_next(CHAR_PENDING):
                self.emit(TokenTransactionPendingFlag, [next(self)])
                return self.state_parsing_payee
            return self.state_parsing_payee
        raise TransactionLexingError("incomplete transaction")

    def state_parsing_payee(self):
        return self.state_parsing_rest_of_line(
            TokenTransactionPayee,
            self.state_parsing_transaction_posting_indentation)

    def state_parsing_rest_of_line(
        self,
        klass, next_state,
        allow_empty_values=False
    ):
        chars = []
        while self.more():
            if self.peek() in CHAR_ENTER:
                next(self)
                while chars and chars[-1] in CHAR_WHITESPACE:
                    chars = chars[:-1]
                break
            if self.peek() in CHAR_WHITESPACE and not chars:
                next(self)
                continue
            chars.append(next(self))
        if allow_empty_values or chars:
            self.emit(klass, chars)
            return next_state
        raise TransactionLexingError("incomplete transaction")

    def state_parsing_transaction_posting_indentation(self):
        chars = []
        while self.more():
            if self.peek() not in CHAR_WHITESPACE:
                break
            chars.append(next(self))
        if not chars:
            return
        if self.more() and self.peek() in CHAR_ENTER:
            next(self)
            return self.state_parsing_transaction_posting_indentation
        return self.state_parsing_transaction_posting_account

    def state_parsing_transaction_comment(self):
        return self.state_parsing_rest_of_line(
            TokenTransactionComment,
            self.state_parsing_transaction_posting_indentation)

    def state_parsing_transaction_posting_account(self):
        chars = []
        if self.more() and self.peek() in CHAR_COMMENT:
            return self.state_parsing_transaction_comment
        while self.more():
            if (
                (self.peek() in CHAR_WHITESPACE and
                 chars and chars[-1] in CHAR_WHITESPACE) or
                self.peek() in CHAR_TAB or
                self.peek() in CHAR_ENTER
            ):
                while chars[-1] in CHAR_WHITESPACE:
                    chars = chars[:-1]
                break
            chars.append(next(self))
        if not chars:
            raise TransactionLexingError("truncated transaction posting")
        self.emit(TokenTransactionPostingAccount, chars)
        return self.state_parsing_transaction_posting_amount

    def state_parsing_transaction_posting_amount(self):
        return self.state_parsing_rest_of_line(
            TokenTransactionPostingAmount,
            self.state_parsing_transaction_posting_indentation,
            allow_empty_values=True)

    def run(self):
        state = self.state_parsing_transaction_date
        while state:
            state = state()


class LedgerContextualLexer(GenericLexer):

    def state_parsing_toplevel(self):
        while self.more():
            if isinstance(self.peek(), TokenComment):
                return self.state_parsing_comment
            token = next(self)
            self.emit(token.__class__, token.contents)

    def state_parsing_comment(self):
        token = next(self)
        if (
            self.more() and
            isinstance(token, TokenComment) and
            isinstance(self.peek(), TokenTransaction)
        ):
            transaction_token = next(self)
            additional_comments = []
            while self.more() and isinstance(self.peek(), TokenComment):
                additional_comments.append(next(self))
            self.emit(TokenTransactionWithContext,
                      [token, transaction_token] + additional_comments)
        else:
            self.emit(token.__class__, token.contents)
        return self.state_parsing_toplevel

    def run(self):
        state = self.state_parsing_toplevel
        while state:
            try:
                state = state()
            except LexingError:
                raise
            except Exception as e:
                raise LexingError(
                    "error parsing ledger data between chunk %d and chunk %d): %s" % (
                        self._last_emitted_pos, self.pos, e
                    )
                )


def lex_ledger_file_contents(text, debug=False):
    lexer = LedgerTextLexer(text)
    lexer.run()
    concat_lexed = "".join([x.contents for x in lexer.tokens])
    if concat_lexed != text:
        if debug:
            u = "Debugging error lexing text: files differ\n\n"
            diffing.two_way_diff(u + text, u + concat_lexed)
        raise LexingError("the lexed contents and the original contents are not the same")
    lexer = LedgerContextualLexer(lexer.tokens)
    lexer.run()
    concat_lexed = "".join([ x.contents for x in lexer.tokens ])
    if concat_lexed != text:
        if debug:
            u = "Debugging error lexing chunks: files differ\n\n"
            diffing.two_way_diff(u + text, u + concat_lexed)
        raise LexingError("the lexed chunks and the original chunks are not the same")
    return lexer.tokens



#!/usr/bin/python3

import pickle
import calendar
import codecs
import collections
import datetime
import fcntl
import fnmatch
import logging
import re
import os
import signal
import struct
import sys
import termios
import threading
import time
import tty

__version__ = "0.3.8"


log = logging.getLogger(__name__)


def debug(string, *args):
    log.debug(string, *args)


_debug_time = False


def debug_time(logger):
    def debug_time_inner(kallable):
        def f(*a, **kw):
            global _debug_time
            if not _debug_time:
                return kallable(*a, **kw)
            start = time.time()
            try:
                name = kallable.__name__
            except AttributeError:
                name = str(kallable)
            name = name + "@" + threading.currentThread().getName()
            try:
                logger.debug("* Timing:    %-55s  started", name)
                return kallable(*a, **kw)
            finally:
                end = time.time() - start
                logger.debug("* Timed:     %-55s  %.3f seconds", name, end)
        return f
    return debug_time_inner


def enable_debugging(enable):
    global _debug_time
    if enable:
        _debug_time = True
        fmt = "%(created)f:%(levelname)8s:%(name)20s: %(message)s"
        logging.basicConfig(level=logging.DEBUG, format=fmt)


def matches(string, options):
    """Returns True if the string case-insensitively glob-matches any of the
    globs present in options."""
    for option in options:
        if fnmatch.fnmatch(string, option):
            return True
    return False


class LedgerConfigurationError(Exception):
    pass


class TransactionInputValidationError(ValueError):
    pass


def find_ledger_file(ledger_file=None):
    """Returns main ledger file path or raise exception if it cannot be \
found.  If ledger_file is not None, use that path."""
    if ledger_file is not None:
        return os.path.abspath(ledger_file)
    ledgerrcpath = os.path.abspath(os.path.expanduser("~/.ledgerrc"))
    if "LEDGER_FILE" in os.environ:
        return os.path.abspath(os.path.expanduser(os.environ["LEDGER_FILE"]))
    elif os.path.exists(ledgerrcpath):
        # hacky
        ledgerrc = open(ledgerrcpath).readlines()
        pat = r"^--file\s+(.*?)\s*$"
        matches = [ re.match(pat, m) for m in ledgerrc ]
        matches = [ m.group(1) for m in matches if m ]
        if not matches:
            raise LedgerConfigurationError("LEDGER_FILE environment variable not set, and your .ledgerrc file does not contain a --file parameter.")
        return os.path.abspath(os.path.expanduser(matches[0]))
    else:
        raise LedgerConfigurationError("LEDGER_FILE environment variable not set, and no \
.ledgerrc file found.")


def add_months(sourcedate, months):
    month = sourcedate.month - 1 + months
    year = int(sourcedate.year + month / 12 )
    month = month % 12 + 1
    day = min(sourcedate.day,calendar.monthrange(year,month)[1])
    return datetime.date(year,month,day)


def find_ledger_price_file(price_file=None):
    """Returns main ledger file path or raise exception if it cannot be \
found.  If price_file is not None, use that path."""
    if price_file is not None:
        return os.path.abspath(price_file)
    ledgerrcpath = os.path.abspath(os.path.expanduser("~/.ledgerrc"))
    if "LEDGER_PRICE_DB" in os.environ:
        return os.path.abspath(os.path.expanduser(os.environ["LEDGER_PRICE_DB"]))
    elif os.path.exists(ledgerrcpath):
        # hacky
        ledgerrc = open(ledgerrcpath).readlines()
        pat = r"^--price-db\s+(.+)"
        matches = [ re.match(pat, m) for m in ledgerrc ]
        matches = [ m.group(1) for m in matches if m ]
        if not matches:
            raise LedgerConfigurationError("LEDGER_PRICE_DB environment variable not set, and your .ledgerrc file does not contain a --price-db parameter.")
        return os.path.abspath(os.path.expanduser(matches[0]))
    else:
        raise LedgerConfigurationError("LEDGER_PRICE_DB environment variable not set, and no \
.ledgerrc file found.")


def generate_price_records(records):
    """Generates a set of price records.

    records is a list containing tuples.  each tuple contains:
      commodity is a ledger commodity
      price is the price in ledger.Amount form
      date is a datetime.date
    """
    lines = [""]
    longestcomm = max(list(len(str(a[0])) for a in records))
    longestamount = max(list(len(str(a[1])) for a in records))
    for commodity, price, date in records:
        fmt = "P %s %-" + str(longestcomm) + "s %" + str(longestamount) + "s"
        lines.append(fmt % (
            date.strftime("%Y-%m-%d %H:%M:%S"),
            commodity,
            price,
        ))
    lines.append("")
    return lines


class Settings(dict):

    def __init__(self, filename):
        self.data = dict()
        self.filename = filename

    @classmethod
    def load_or_defaults(cls, filename):
        s = cls(filename)
        if os.path.isfile(s.filename):
            try:
                s.data = pickle.load(open(s.filename, "rb"))
            except Exception as e:
                log.error("Cannot load %s so loading defaults: %s", s.filename, e)
        try:
            unused_suggester = s["suggester"]
        except KeyError:
            s["suggester"] = AccountSuggester()
        return s

    def __setitem__(self, item, value):
        self.data[item] = value
        self.persist()

    def __getitem__(self, item):
        return self.data[item]

    def __delitem__(self, item):
        if item in self.data:
            del self.data[item]
            self.persist()

    def keys(self):
        return list(self.data.keys())

    def get(self, item, default):
        return self.data.get(item, default)

    def persist(self):
        p = open(self.filename, "wb")
        pickle.dump(self.data, p)
        p.flush()
        p.close()


class AccountSuggester(object):

    def __init__(self):
        self.account_to_words = dict()

    def __str__(self):
        dump = str(self.account_to_words)
        return "<AccountSuggester %s>" % dump

    def associate(self, words, account):
        words = [ w.lower() for w in words.split() ]
        account = str(account)
        if account not in self.account_to_words:
            self.account_to_words[account] = dict()
        for w in words:
            if w not in self.account_to_words[account]:
                self.account_to_words[account][w] = 0
            self.account_to_words[account][w] += 1

    def suggest(self, words):
        words = [ w.lower() for w in words.split() ]
        account_counts = dict()
        for account, ws in list(self.account_to_words.items()):
            for w, c in list(ws.items()):
                if w in words:
                    if not account in account_counts:
                        account_counts[account] = 0
                    account_counts[account] += c
        results = list(reversed(sorted(
            list(account_counts.items()), key=lambda x: x[1]
        )))
        if results:
            return results[0][0]
        return None


# ======================  GTK =======================


TransactionPosting = collections.namedtuple(
    'TransactionPosting',
    ['account', 'amount']
)



import ledger
import ledgerhelpers
import ledgerhelpers.legacy as hl
import ledgerhelpers.legacy_needsledger as hln
import os
import sys
import threading

import gi
gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
from gi.repository import GObject
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango


EVENT_TAB = 65289
EVENT_SHIFTTAB = 65056
EVENT_ESCAPE = 65307


_css_adjusted = {}


def g_async(func, success_func, failure_func):
    def f():
        try:
            GObject.idle_add(success_func, func())
        except BaseException as e:
            GObject.idle_add(failure_func, e)
    t = threading.Thread(target=f)
    t.setDaemon(True)
    t.start()
    return t


def add_css(css):
    # Must only ever be called at runtime, not at import time.
    global _css_adjusted
    if css not in _css_adjusted:
        style_provider = Gtk.CssProvider()
        if isinstance(css, str):
            css = css.encode("utf-8")
        style_provider.load_from_data(css)
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(),
            style_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )
    _css_adjusted[css] = True


def FatalError(message, secondary=None, parent=None):
    d = Gtk.MessageDialog(
        parent,
        Gtk.DialogFlags.DESTROY_WITH_PARENT,
        Gtk.MessageType.ERROR,
        Gtk.ButtonsType.CLOSE,
        message,
    )
    if secondary:
        d.format_secondary_text(secondary)
    d.run()


def cannot_start_dialog(msg):
    return FatalError("Cannot start program", msg)


class EagerCompletion(Gtk.EntryCompletion):
    """Completion class that substring matches within a builtin ListStore."""

    def __init__(self, *args):
        Gtk.EntryCompletion.__init__(self, *args)
        self.set_model(Gtk.ListStore(GObject.TYPE_STRING))
        self.set_match_func(self.iter_points_to_matching_entry)
        self.set_text_column(0)
        self.set_inline_completion(True)

    def iter_points_to_matching_entry(self, unused_c, k, i, unused=None):
        model = self.get_model()
        acc = model.get(i, 0)[0].lower()
        if k.lower() in acc:
            return True
        return False


def load_journal_and_settings_for_gui(price_file_mandatory=False,
                                      ledger_file=None,
                                      price_file=None):
    try:
        ledger_file = ledgerhelpers.find_ledger_file(ledger_file)
    except Exception as e:
        cannot_start_dialog(str(e))
        sys.exit(4)
    try:
        price_file = ledgerhelpers.find_ledger_price_file(price_file)
    except ledgerhelpers.LedgerConfigurationError as e:
        if price_file_mandatory:
            cannot_start_dialog(str(e))
            sys.exit(4)
        else:
            price_file = None
    except Exception as e:
        cannot_start_dialog(str(e))
        sys.exit(4)
    try:
        from ledgerhelpers.journal import Journal
        journal = Journal.from_file(ledger_file, price_file)
    except Exception as e:
        cannot_start_dialog("Cannot open ledger file: %s" % e)
        sys.exit(5)
    s = ledgerhelpers.Settings.load_or_defaults(
        os.path.expanduser("~/.ledgerhelpers.ini")
    )
    return journal, s


def find_ledger_file_for_gui():
    try:
        ledger_file = ledgerhelpers.find_ledger_file()
        return ledger_file
    except Exception as e:
        cannot_start_dialog(str(e))
        sys.exit(4)


class EagerCompletingEntry(Gtk.Entry):
    """Entry that substring-matches eagerly using a builtin ListStore-based
    Completion, and also accepts defaults.
    """

    prevent_completion = False

    def __init__(self, *args):
        Gtk.Entry.__init__(self, *args)
        self.default_text = ''
        self.old_default_text = ''
        self.set_completion(EagerCompletion())

    def set_default_text(self, default_text):
        self.old_default_text = self.default_text
        self.default_text = default_text
        if not self.get_text() or self.get_text() == self.old_default_text:
            self.set_text(self.default_text)


class LedgerAmountEntry(Gtk.Grid):

    __gsignals__ = {
        'changed': (GObject.SIGNAL_RUN_LAST, None, ())
    }
    default_commodity = None

    def show(self):
        Gtk.Grid.show(self)
        self.entry.show()
        if self.display:
            self.display.show()

    def do_changed(self):
        pass

    def set_placeholder_text(self, text):
        self.entry.set_placeholder_text(text)

    def __init__(self, display=True):
        Gtk.Grid.__init__(self)
        self.amount = None
        self.entry = Gtk.Entry()
        self.entry.set_width_chars(8)
        if display:
            self.display = Gtk.Label()
        else:
            self.display = None
        self.entry.set_alignment(1.0)
        self.attach(self.entry, 0, 0, 1, 1)
        if self.display:
            self.attach(self.display, 1, 0, 1, 1)
            self.display.set_halign(Gtk.Align.END)
            self.display.set_justify(Gtk.Justification.RIGHT)
        self.set_column_spacing(4)
        self.donotreact = False
        self.entry.connect("changed", self.entry_changed)
        self.set_default_commodity(ledger.Amount("$ 1").commodity)
        self.set_activates_default = self.entry.set_activates_default

    def get_default_commodity(self):
        return self.default_commodity

    def set_default_commodity(self, commodity):
        if isinstance(commodity, ledger.Amount):
            commodity = commodity.commodity
        if str(self.default_commodity) != str(commodity):
            self.default_commodity = commodity
        self.entry.emit("changed")

    def is_focus(self):
        return self.entry.is_focus()

    def grab_focus(self):
        self.entry.grab_focus()

    def get_amount(self):
        return self.amount

    def set_amount(self, amount, skip_entry_update=False):
        self.amount = amount
        if self.display:
            self.display.set_text(str(amount) if amount is not None else "")
        self.donotreact = True
        if not skip_entry_update:
            self.entry.set_text(str(amount) if amount is not None else "")
        self.donotreact = False
        self.emit("changed")

    def set_text(self, text):
        self.entry.set_text(text)

    def _adjust_entry_size(self, w):
        text = w.get_text()
        w.set_width_chars(max([8, len(text)]))

    def entry_changed(self, w, *unused_args):
        self._adjust_entry_size(w)

        if self.donotreact:
            return

        text = self.entry.get_text()

        try:
            p = ledger.Amount(text)
        except ArithmeticError:
            self.set_amount(None, True)
            self.emit("changed")
            return

        if not str(p.commodity):
            p.commodity = self.default_commodity
        if str(p):
            self.set_amount(p, True)
        else:
            self.set_amount(None, True)

        self.emit("changed")


class LedgerAmountWithPriceEntry(LedgerAmountEntry):

    def __init__(self, display=True):
        self.price = None
        LedgerAmountEntry.__init__(self, display=display)

    def get_amount_and_price(self):
        return self.amount, self.price

    def get_amount_and_price_formatted(self):
        if self.amount and self.price:
            return str(self.amount) + " " + self.price.strip()
        elif self.amount:
            return str(self.amount)
        elif self.price:
            return self.price
        else:
            return ""

    def set_amount_and_price(self, amount, price, skip_entry_update=False):
        self.amount = amount
        self.price = price
        p = [
            str(amount if amount is not None else "").strip(),
            str(price if price is not None else "").strip(),
        ]
        p = [x for x in p if x]
        concat = " ".join(p)
        if self.display:
            self.display.set_text(concat)
        self.donotreact = True
        if not skip_entry_update:
            self.entry.set_text(concat)
        self.donotreact = False

    def entry_changed(self, w, *unused_args):
        self._adjust_entry_size(w)

        if self.donotreact:
            return

        text = self.entry.get_text()
        i = text.find("@")
        if i != -1:
            price = text[i:] if text[i:] else None
            text = text[:i]
        else:
            price = None

        try:
            p = ledger.Amount(text)
        except ArithmeticError:
            self.set_amount_and_price(None, price, True)
            self.emit("changed")
            return

        if not str(p.commodity):
            p.commodity = self.default_commodity
        if str(p):
            self.set_amount_and_price(p, price, True)
        else:
            self.set_amount_and_price(None, price, True)

        self.emit("changed")


class EditableTabFocusFriendlyTextView(Gtk.TextView):

    def __init__(self, *args):
        Gtk.TextView.__init__(self, *args)
        self.connect("key-press-event", self.handle_tab)

    def handle_tab(self, widget, event):
        if event.keyval == EVENT_TAB:
            widget.get_toplevel().child_focus(Gtk.DirectionType.TAB_FORWARD)
            return True
        elif event.keyval == EVENT_SHIFTTAB:
            widget.get_toplevel().child_focus(Gtk.DirectionType.TAB_BACKWARD)
            return True
        return False


class LedgerTransactionView(Gtk.Box):

    css = b"""
ledgertransactionview {
  border: 1px @borders inset;
}
"""

    def __init__(self, *args):
        add_css(self.css)
        Gtk.Box.__init__(self)
        self.textview = EditableTabFocusFriendlyTextView(*args)
        self.textview.override_font(
            Pango.font_description_from_string('monospace')
        )
        self.textview.set_border_width(12)
        self.textview.set_hexpand(True)
        self.textview.set_vexpand(True)
        self.add(self.textview)
        self.textview.get_buffer().set_text(
            "# A live preview will appear here as you input data."
        )

    def get_buffer(self):
        return self.textview.get_buffer()

    def generate_record(self, *args):
        lines = hln.generate_record(*args)
        self.textview.get_buffer().set_text("\n".join(lines))


LedgerTransactionView.set_css_name("ledgertransactionview")


class EscapeHandlingMixin(object):

    escape_handling_suspended = False

    def activate_escape_handling(self):
        self.connect("key-press-event", self.handle_escape)

    def suspend_escape_handling(self):
        self.escape_handling_suspended = True

    def resume_escape_handling(self):
        self.escape_handling_suspended = False

    def handle_escape(self, unused_window, event, unused_user_data=None):
        if (
            not self.escape_handling_suspended and
            event.keyval == EVENT_ESCAPE
        ):
            self.emit('delete-event', None)
            return True
        return False



import datetime
import ledgerhelpers as m
import ledgerhelpers.legacy as mc
try:
    import ledgerhelpers.journal as journal
except ImportError:
    journal = None
import tests.test_base as base
import tempfile
import unittest
from unittest import TestCase as T


@unittest.skipIf(journal is None, reason="ledger-python is not available on this system")
class TestJournal(T):

    def test_journal_with_simple_transaction(self):
        c = base.datapath("simple_transaction.dat")
        j = journal.Journal.from_file(c, None)
        payees = j.all_payees()
        self.assertListEqual(payees, ["beer"])
        accts, commos = j.accounts_and_last_commodity_for_account()
        expaccts = ["Accounts:Cash", "Expenses:Drinking"]
        self.assertListEqual(accts, expaccts)
        self.assertEqual(commos["Expenses:Drinking"], "1.00 CHF")

    def test_reload_works(self):
        with tempfile.NamedTemporaryFile(mode="w") as f:
            with open(base.datapath("simple_transaction.dat")) as transaction_data:
                data = transaction_data.read()
            f.write(data)
            f.flush()
            j = journal.Journal.from_file(f.name, None)
            _, commos = j.accounts_and_last_commodity_for_account()
            self.assertEqual(commos["Expenses:Drinking"], "1.00 CHF")
            data = data.replace("CHF", "EUR")
            f.write(data)
            f.flush()
            _, commos = j.accounts_and_last_commodity_for_account()
            self.assertEqual(commos["Expenses:Drinking"], "1.00 EUR")

    def test_transactions_with_payee_match(self):
        c = base.datapath("simple_transaction.dat")
        j = journal.Journal.from_file(c, None)
        ts = journal.transactions_with_payee("beer", j.internal_parsing())
        self.assertEqual(ts[0].payee, "beer")


class TestGenerateRecord(T):

    def test_no_spurious_whitespace(self):
        title = "x"
        date = datetime.date(2014, 1, 1)
        cleared_date = None
        accountamounts = [
            ("assets", "56 CHF"),
            ("expenses", ""),
        ]
        res = mc.generate_record(title, date, cleared_date, "", accountamounts)
        self.assertListEqual(
            res,
            """
2014-01-01 x
    assets      56 CHF
    expenses

""".splitlines())

    def test_no_cleared_date_when_cleared_date_not_supplied(self):
        cases = [
            ("2014-01-01 x", (datetime.date(2014, 1, 1), None), ""),
            ("2014-01-01 * x", (datetime.date(2014, 1, 1), datetime.date(2014, 1, 1)), "*"),
            ("2014-01-01=2015-01-01 ! x", (datetime.date(2014, 1, 1), datetime.date(2015, 1, 1)), "!"),
        ]
        accountamounts = [("assets", "56 CHF"), ("expenses", "")]
        for expected_line, (date, cleared), statechar in cases:
            res = mc.generate_record("x", date, cleared, statechar, accountamounts)[1]
            self.assertEqual(res, expected_line)

    def test_empty_record_auto_goes_last(self):
        accountamounts = [("expenses", ""), ("assets:cash", "56 CHF")]
        res = mc.generate_record("x", datetime.date(2014, 1, 1),
                                None, "", accountamounts)
        self.assertListEqual(
            res,
            """
2014-01-01 x
    assets:cash    56 CHF
    expenses

""".splitlines())



#!/usr/bin/python3

import argparse
import codecs
import collections
import datetime
import itertools
import subprocess
import sys

from ledgerhelpers import diffing
from ledgerhelpers import parser
from ledgerhelpers import gui
from ledgerhelpers.programs import common as common_programs


def get_argparser():
    parser = argparse.ArgumentParser(
        'Sort transactions in a ledger file chronologically',
        parents=[common_programs.get_common_argparser()]
    )
    parser.add_argument('-y', dest='assume_yes', action='store_true',
                        help='record changes immediately, instead of '
                        'showing a three-way diff for you to resolve')
    parser.add_argument('--debug', dest='debug', action='store_true',
                        help='do not capture exceptions into a dialog box')
    return parser


def sort_transactions(items):
    smallest_date = datetime.date(1000, 1, 1)
    largest_date = datetime.date(3000, 1, 1)
    bydates = collections.OrderedDict()
    first_transaction_seen = False
    for n, item in enumerate(items):
        if hasattr(item, "date"):
            first_transaction_seen = True
        if first_transaction_seen:
            later_dates = itertools.chain(
                (getattr(items[i], "date", None) for i in range(n, len(items))),
                [largest_date]
            )
            for date in later_dates:
                if date is not None:
                    break
        else:
            date = smallest_date
        if date not in bydates:
            bydates[date] = []
        bydates[date] += [item]
    for date in sorted(bydates):
        for item in bydates[date]:
            yield item


def main(argv):
    p = get_argparser()
    args = p.parse_args(argv[1:])
    if args.file:
        ledgerfile = args.file
    else:
        ledgerfile = gui.find_ledger_file_for_gui()
    try:
        leftcontents = codecs.open(ledgerfile, "rb", "utf-8").read()
        items = parser.lex_ledger_file_contents(leftcontents, debug=args.debug)
        rightcontents = "".join(i.contents for i in sort_transactions(items))
        if args.assume_yes:
            with open(ledgerfile, "w") as out_file:
                out_file.write(rightcontents)
            return 0
        try:
            diffing.three_way_diff(ledgerfile, leftcontents, rightcontents)
        except subprocess.CalledProcessError as e:
            if args.debug:
                raise
            print("Meld failed", file=sys.stderr)
            print("Meld process failed with return code %s" % e.returncode, file=sys.stderr)
            return e.returncode
    except Exception as e:
        if args.debug:
            raise
        print("Transaction sort failed", file=sys.stderr)
        print("An unexpected error took place:\n%s" % e, file=sys.stderr)
        return 9



#!/usr/bin/python3

import argparse
import collections
import datetime
import http.client
import json
import ledger
import ledgerhelpers
from ledgerhelpers import gui
import threading
import traceback
import urllib.parse
import sys
import yahoo_finance

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import GObject
from gi.repository import Gtk


def get_argparser():
    parser = argparse.ArgumentParser(
        'Update prices in a Ledger price file'
    )
    parser.add_argument('-b', dest='batch', action='store_true',
                        help='update price file in batch (non-GUI) mode')
    parser.add_argument('--debug', dest='debug', action='store_true',
                        help='do not capture exceptions into a dialog box')
    return parser


class QuoteSource(object):
    pass


class DontQuote(QuoteSource):

    def __str__(self):
        return "skip quoting"

    def get_quote(self, commodity, denominated_in):  # @UnusedVariable
        return None, None


class YahooFinanceCommodities(QuoteSource):

    def __str__(self):
        return "Yahoo! Finance commodities"

    def get_quote(
        self,
        commodity,
        denominated_in,
        commodity_is_currency_pair=False
    ):
        """Returns the price in the appraised_as currency, and the datetime.

        Args:
            commodity: a Ledger commodity representing a non-currency
                       commodity
            denominated_in: a Ledger commodity

        Returns:
            price: ledger.Amount instance
            datetime: datetime.datetime instance
        """
        if not isinstance(commodity, ledger.Commodity):
            raise ValueError("commodity must be a Ledger commodity")
        if not isinstance(denominated_in, ledger.Commodity):
            raise ValueError("denominated_in must be a Ledger commodity")
        if commodity_is_currency_pair:
            source = str(commodity)
            source = source if source != "$" else "USD"
            target = str(denominated_in)
            target = target if target != "$" else "USD"
            pair = source + target
            s = yahoo_finance.Currency(pair)
            try:
                price, date = s.get_rate(), s.get_trade_datetime()
            except KeyError:
                raise ValueError(
                    "Yahoo! Finance can't find currency pair %s" % pair
                )
        else:
            if str(denominated_in) not in ["$", "USD"]:
                raise ValueError(
                    "Yahoo! Finance can't quote in %s" % denominated_in
                )
            s = yahoo_finance.Share(str(commodity))
            try:
                price, date = s.get_price(), s.get_trade_datetime()
            except KeyError:
                raise ValueError(
                    "Yahoo! Finance can't find commodity %s" % commodity
                )
        a = ledger.Amount(price)
        a.commodity = denominated_in
        d = datetime.datetime.strptime(
            date,
            '%Y-%m-%d %H:%M:%S UTC+0000'
        )
        return a, d


class YahooFinanceCurrencies(YahooFinanceCommodities):

    def __str__(self):
        return "Yahoo! Finance currencies"

    def get_quote(self, commodity, denominated_in="$"):
        """Returns the price in the appraised_as currency, and the datetime.

        Args:
            commodity: a Ledger commodity representing a currency
            denominated_in: a Ledger commodity

        Returns:
            price: ledger.Amount instance
            datetime: datetime.datetime instance
        """
        return YahooFinanceCommodities.get_quote(
            self,
            commodity,
            denominated_in,
            commodity_is_currency_pair=True
        )


def json_from_uri(uri):
    c = http.client.HTTPSConnection(urllib.parse.urlsplit(uri).netloc)  # @UndefinedVariable
    c.request('GET', uri)
    response = c.getresponse().read()
    try:
        return json.loads(response)
    except ValueError:
        raise ValueError("JSON object undecodable: %s" % response)


class BitcoinCharts(QuoteSource):

    def __str__(self):
        return "bitcoin charts"

    def get_quote(self, commodity, denominated_in):
        """Returns the price in the denominated_in currency,
        and the datetime.

        Args:
            commodity: a Ledger commodity
            denominated_in: a Ledger commodity

        Returns:
            price: ledger.Amount instance
            datetime: datetime.datetime instance
        """
        if not isinstance(commodity, ledger.Commodity):
            raise ValueError("commodity must be a Ledger commodity")
        if not isinstance(denominated_in, ledger.Commodity):
            raise ValueError("denominated_in must be a Ledger commodity")
        if str(commodity) not in ["BTC", "XBT"]:
            raise ValueError(
                "bitcoin charts can only provide quotes for BTC / XBT"
            )

        data = json_from_uri(
            "https://api.bitcoincharts.com/v1/weighted_prices.json"
        )
        try:
            k = "USD" if str(denominated_in) == "$" else str(denominated_in)
            amount = data[k].get("24h", data[k]["7d"])
        except KeyError:
            raise ValueError(
                "bitcoin charts can't provide quotes in %s" % denominated_in
            )
        a = ledger.Amount(amount)
        a.commodity = denominated_in
        d = datetime.datetime.now()
        return a, d


class PriceGatheringDatabase(Gtk.ListStore):

    def __init__(self):
        # Columns:
        # 0: Ledger.commodity to appraise
        # 1: data source to fetch prices from
        # 2: list of (Ledger.commodity) representing which denominations
        #    the (0) commodity must be appraised in
        # 3: list of (gathered price, datetime)
        # 4: errors (exceptions)
        Gtk.ListStore.__init__(self, object, object, object, object, object)
        self.commodities_added = dict()

    def add_to_gather_list(self, commodity, datasource, fetch_prices_in):
        if self.commodities_added.get(str(commodity)):
            return
        if isinstance(commodity, ledger.Amount):
            commodity = commodity.commodity
        assert isinstance(commodity, ledger.Commodity), commodity
        assert isinstance(datasource, QuoteSource)
        for f in fetch_prices_in:
            assert isinstance(f, ledger.Commodity)
        self.append((commodity, datasource, list(fetch_prices_in),
                     list(), list()))
        self.commodities_added[str(commodity)] = True

    def clear_gathered(self):
        for row in self:
            while row[3]:
                row[3].pop()
            while row[4]:
                row[4].pop()
            self.emit('row-changed', row.path, row.iter)

    def record_gathered(self, commodity, amount, timeobject):
        assert isinstance(commodity, ledger.Commodity), commodity
        assert isinstance(amount, ledger.Amount), amount
        found = False
        for row in self:
            if commodity == row[0]:
                found = True
                break
        assert found, "%s not found in gather list" % commodity
        row[3].append((amount, timeobject))
        self.emit('row-changed', row.path, row.iter)

    def record_gathered_error(self, commodity, error):
        assert isinstance(commodity, ledger.Commodity), commodity
        found = False
        for row in self:
            if commodity == row[0]:
                found = True
                break
        assert found, "%s not found in gather list" % commodity
        row[4].append(error)
        self.emit('row-changed', row.path, row.iter)

    def get_currency_by_path(self, treepath):
        it = self.get_iter(treepath)
        return self.get_value(it, 0)

    def update_quoter(self, treepath, new_datasource):
        i = self.get_iter(treepath)
        self.set_value(i, 1, new_datasource)

    def update_price_in(self, treepath, new_price_in):
        i = self.get_iter(treepath)
        self.set_value(i, 2, new_price_in)

    def get_prices(self):
        for row in self:
            for p, d in row[3]:
                yield row[0], p, d

    def get_errors(self):
        for row in self:
            for e in row[4]:
                yield row[0], e


@GObject.type_register
class PriceGatherer(GObject.GObject):

    __gsignals__ = {
        "gathering-started": (
             GObject.SIGNAL_RUN_LAST, None, ()
        ),
        "gathering-done": (
             GObject.SIGNAL_RUN_LAST, None, ()
        ),
    }

    def __init__(self, quoters):
        GObject.GObject.__init__(self)
        assert quoters
        self.quoters = quoters
        self.database = PriceGatheringDatabase()

    def load_commodities_from_journal(
        self,
        journal,
        map_from_currencystrs_to_quotesources,
        map_from_currencystrs_to_priceins,
    ):
        DontQuote = self.quoters.get("DontQuote", list(self.quoters.values())[0])
        default = "$"
        coms = [c.strip_annotations().commodity for c in journal.commodities()]
        strcoms = [str(c) for c in coms]
        if "USD" in strcoms and "$" not in strcoms:
            default = "USD"
        already = dict()
        for c in coms:
            if str(c) in already:
                continue
            already[str(c)] = True
            quoter = list(self.quoters.values())[0]
            if str(c) in ["USD", "$"] and default in ["USD", "$"]:
                quoter = DontQuote
            if str(c) in map_from_currencystrs_to_quotesources:
                quoter = map_from_currencystrs_to_quotesources[str(c)]
            quoteins = [journal.commodity(default)]
            if str(c) in map_from_currencystrs_to_priceins:
                quoteins = list(map_from_currencystrs_to_priceins[str(c)])
            self.database.add_to_gather_list(
                c,
                quoter,
                quoteins,
            )

    def _gather_inner(self, sync=False):
        def do(f, *a):
            if not sync:
                return GObject.idle_add(f, *a)
            return f(*a)

        do(self.database.clear_gathered)
        for row in self.database:
            commodity = row[0]
            quotesource = row[1]
            for denominated_in in row[2]:
                try:
                    price, time = quotesource.get_quote(
                        commodity,
                        denominated_in=denominated_in
                    )
                    if price is None and time is None:
                        continue
                    do(
                        self.database.record_gathered,
                        commodity,
                        price,
                        time
                    )
                except Exception as e:
                    error = str(e)
                    do(
                        self.database.record_gathered_error,
                        commodity,
                        error,
                    )
                    traceback.print_exc()
        GObject.idle_add(self.emit, "gathering-done")

    def gather_quotes(self, sync=False):
        GObject.idle_add(self.emit, "gathering-started")
        if not sync:
            t = threading.Thread(target=self._gather_inner)
            t.setDaemon(True)
            t.start()
        else:
            return self._gather_inner(sync=True)


@GObject.type_register
class CellRendererCommodity(Gtk.CellRendererText):

    __gproperties__ = {
        "commodity": (
            GObject.TYPE_PYOBJECT,
            "commodity to display",
            "the commodity to render in the cell",
            GObject.PARAM_READWRITE
        )
    }

    def do_set_property(self, prop, val):
        if prop.name == "commodity":
            self.set_property("text", str(val))
        else:
            self.set_property(prop, val)


@GObject.type_register
class CellRendererQuoteSource(Gtk.CellRendererCombo):

    __gproperties__ = {
        "source": (
            GObject.TYPE_PYOBJECT,
            "source of quotes",
            "the QuoteSource data source to render",
            GObject.PARAM_READWRITE
        )
    }

    def do_set_property(self, prop, val):
        if prop.name == "source":
            self.set_property("text", str(val))
        else:
            self.set_property(prop, val)


@GObject.type_register
class CellRendererPriceIn(Gtk.CellRendererText):

    __gproperties__ = {
        "price-in": (
            GObject.TYPE_PYOBJECT,
            "which commodities to denominate the quotes in",
            "list of denominations for quote fetches",
            GObject.PARAM_READWRITE
        )
    }

    def do_set_property(self, prop, val):
        if prop.name == "price-in":
            self.set_property("text", "\n".join(str(s) for s in val))
        else:
            self.set_property(prop, val)


@GObject.type_register
class CellRendererFetchedList(Gtk.CellRendererText):

    render_datum = 0

    __gproperties__ = {
        "list": (
            GObject.TYPE_PYOBJECT,
            "list to display",
            "the list of fetched quotes to render",
            GObject.PARAM_READWRITE
        ),
        "render-datum": (
            GObject.TYPE_INT,
            "which datum to render",
            "render either 0 for price or 1 for time",
            render_datum,
            1,
            0,
            GObject.PARAM_READWRITE
        ),
    }

    def do_get_property(self, prop):
        if prop.name == "render-datum":
            return self.render_datum
        else:
            assert 0, prop

    def do_set_property(self, prop, val):
        def render(obj):
            if self.render_datum != 1:
                return str(obj)
            else:
                return obj.strftime("%Y-%m-%d %H:%M:%S")
        if prop.name == "list":
            self.set_property(
                "text",
                "\n".join(render(p[self.render_datum]) for p in val)
            )
        elif prop.name == "render-datum":
            self.render_datum = val
        else:
            assert 0, (prop, val)


@GObject.type_register
class CellRendererFetchErrors(Gtk.CellRendererText):

    __gproperties__ = {
        "errors": (
            GObject.TYPE_PYOBJECT,
            "list of errors to render",
            "list of error strings to render",
            GObject.PARAM_READWRITE
        )
    }

    def do_set_property(self, prop, val):
        if prop.name == "errors":
            self.set_property("text", "\n".join(str(s) for s in val))
        else:
            self.set_property(prop, val)


class PriceGatheringView(Gtk.TreeView):

    def __init__(self):
        Gtk.TreeView.__init__(self)
        commodity_renderer = CellRendererCommodity()
        commodity_renderer.set_property("yalign", 0.0)
        commodity_column = Gtk.TreeViewColumn(
            "Commodity",
            commodity_renderer,
            commodity=0,
        )
        self.append_column(commodity_column)

        quotesource_renderer = CellRendererQuoteSource()
        quotesource_renderer.set_property("yalign", 0.0)
        quotesource_column = Gtk.TreeViewColumn(
            "Quote source",
            quotesource_renderer,
            source=1,
        )
        self.append_column(quotesource_column)
        self.quotesource_renderer = quotesource_renderer

        price_in_renderer = CellRendererPriceIn()
        price_in_renderer.set_property("yalign", 0.0)
        price_in_column = Gtk.TreeViewColumn(
            "Quote in",
            price_in_renderer,
            price_in=2,
        )
        self.append_column(price_in_column)
        self.price_in_renderer = price_in_renderer

        fetched_prices_renderer = CellRendererFetchedList()
        fetched_prices_renderer.set_property("yalign", 0.0)
        fetched_prices_renderer.set_property("render-datum", 0)
        fetched_prices_column = Gtk.TreeViewColumn(
            "Price",
            fetched_prices_renderer,
            list=3,
        )
        self.append_column(fetched_prices_column)

        fetched_dates_renderer = CellRendererFetchedList()
        fetched_dates_renderer.set_property("yalign", 0.0)
        fetched_dates_renderer.set_property("render-datum", 1)
        fetched_dates_column = Gtk.TreeViewColumn(
            "Date",
            fetched_dates_renderer,
            list=3,
        )
        self.append_column(fetched_dates_column)

        errors_renderer = CellRendererFetchErrors()
        errors_renderer.set_property("yalign", 0.0)
        errors_column = Gtk.TreeViewColumn(
            "Status",
            errors_renderer,
            errors=4,
        )
        self.append_column(errors_column)


class UpdatePricesWindow(Gtk.Window):

    def __init__(self):
        Gtk.Window.__init__(self, title="Update prices")
        self.set_border_width(12)
        self.set_default_size(800, 430)

        grid = Gtk.Grid()
        grid.set_column_spacing(8)
        grid.set_row_spacing(8)
        self.add(grid)

        row = 0

        self.gatherer_view = PriceGatheringView()
        self.gatherer_view.set_hexpand(True)
        self.gatherer_view.set_vexpand(True)

        sw = Gtk.ScrolledWindow()
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sw.set_shadow_type(Gtk.ShadowType.IN)
        sw.add(self.gatherer_view)
        grid.attach(sw, 0, row, 1, 1)

        row += 1

        button_box = Gtk.ButtonBox()
        button_box.set_layout(Gtk.ButtonBoxStyle.END)
        button_box.set_spacing(12)
        self.status = Gtk.Label()
        button_box.add(self.status)
        self.close_button = Gtk.Button(stock=Gtk.STOCK_CLOSE)
        button_box.add(self.close_button)
        self.fetch_button = Gtk.Button(label="Fetch")
        button_box.add(self.fetch_button)
        self.save_button = Gtk.Button(stock=Gtk.STOCK_SAVE)
        button_box.add(self.save_button)
        grid.attach(button_box, 0, row, 2, 1)
        self.fetch_button.set_can_default(True)
        self.fetch_button.grab_default()


class UpdatePricesCommon(object):

    def __init__(self, journal, preferences):
        self.journal = journal
        self.preferences = preferences
        try:
            self.preferences["quotesources"]
        except KeyError:
            self.preferences["quotesources"] = dict()
        try:
            self.preferences["quotecurrencies"]
        except KeyError:
            self.preferences["quotecurrencies"] = dict()
        self.quoters = collections.OrderedDict(
            (str(q), q) for q in [
                YahooFinanceCommodities(),
                YahooFinanceCurrencies(),
                BitcoinCharts(),
                DontQuote(),
            ]
        )
        self.gatherer = PriceGatherer(self.quoters)

    def get_ready(self):
        prefquotesources = dict(
            (cur, self.quoters.get(n, list(self.quoters.values())[0]))
            for cur, n
            in list(self.preferences["quotesources"].items())
            if type(n) is not list
        )
        prefquotecurrencies = dict(
            (cur, [self.journal.commodity(v, True) for v in pins])
            for cur, pins
            in list(self.preferences["quotecurrencies"].items())
        )
        self.gatherer.load_commodities_from_journal(
            self.journal,
            prefquotesources,
            prefquotecurrencies,
        )

    def save_fetched_prices(self):
        recs = list(self.gatherer.database.get_prices())
        if recs:
            lines = self.journal.generate_price_records(recs)
            self.journal.add_text_to_price_file(lines)

    def output_errors(self):
        recs = list(self.gatherer.database.get_errors())
        if recs:
            print("There were errors obtaining prices:", file=sys.stderr)
        for comm, error in recs:
            print("* Gathering %s: %s" % (comm, error))
        return bool(recs)

    def run(self):
        self.get_ready()
        self.gatherer.gather_quotes(sync=True)
        errors = self.output_errors()
        self.save_fetched_prices()
        if errors:
            return 3


class UpdatePricesApp(
    UpdatePricesCommon,
    UpdatePricesWindow,
    gui.EscapeHandlingMixin
):

    def __init__(self, journal, preferences):
        UpdatePricesCommon.__init__(self, journal, preferences)
        UpdatePricesWindow.__init__(self)

        self.fetch_level = 0
        self.connect("delete-event", lambda *unused_a: self.save_preferences())
        self.activate_escape_handling()

        self.gatherer_view.set_model(self.gatherer.database)
        self.close_button.connect(
            "clicked",
            lambda _: self.emit('delete-event', None)
        )
        # FIXME: do asynchronously.
        self.get_ready()

    def get_ready(self):
        UpdatePricesCommon.get_ready(self)

        quotesource_model = Gtk.ListStore(str, object)
        for q in list(self.quoters.items()):
            quotesource_model.append(q)
        self.gatherer_view.quotesource_renderer.set_property(
            "model",
            quotesource_model
        )
        self.gatherer_view.quotesource_renderer.set_property(
            "text-column",
            0
        )
        self.gatherer_view.quotesource_renderer.connect(
            "editing-started",
            self.on_quotesource_editing_started
        )
        self.gatherer_view.quotesource_renderer.connect(
            "edited",
            self.on_quotesource_editing_done
        )
        self.gatherer_view.quotesource_renderer.connect(
            "editing-canceled",
            self.on_quotesource_editing_canceled
        )
        self.gatherer_view.price_in_renderer.connect(
            "editing-started",
            self.on_price_in_editing_started
        )
        self.gatherer_view.price_in_renderer.connect(
            "edited",
            self.on_price_in_editing_done
        )
        self.gatherer_view.price_in_renderer.connect(
            "editing-canceled",
            self.on_price_in_editing_canceled
        )
        self.save_button.connect("clicked", self.save_fetched_prices)
        self.fetch_button.connect("clicked", lambda _: self.do_fetch())
        self.gatherer.connect("gathering-started", self.prevent_fetch)
        self.gatherer.connect("gathering-started", self.disallow_save)
        self.gatherer.connect("gathering-started", self.disable_cell_editing)
        self.gatherer.connect("gathering-done", self.enable_cell_editing)
        self.gatherer.connect("gathering-done", self.allow_fetch)
        self.gatherer.connect("gathering-done", self.allow_save)
        self.gatherer.connect("gathering-done", self.focus_save)
        self.allow_fetch()
        self.disallow_save()
        self.enable_cell_editing()

    def enable_cell_editing(self, unused_w=None):
        self.gatherer_view.quotesource_renderer.set_property(
            "editable",
            True
        )
        self.gatherer_view.price_in_renderer.set_property(
            "editable",
            True
        )

    def disable_cell_editing(self, unused_w=None):
        self.gatherer_view.quotesource_renderer.set_property(
            "editable",
            False
        )
        self.gatherer_view.price_in_renderer.set_property(
            "editable",
            False
        )

    def disallow_save(self, unused_w=None):
        self.save_button.set_sensitive(False)

    def allow_save(self, unused_w=None):
        self.save_button.set_sensitive(True)

    def focus_save(self, unused_w=None):
        self.save_button.grab_focus()

    def prevent_fetch(self, unused_w=None):
        self.fetch_level -= 1
        self.fetch_button.set_sensitive(self.fetch_level > 0)

    def allow_fetch(self, unused_w=None):
        self.fetch_level += 1
        self.fetch_button.set_sensitive(self.fetch_level > 0)

    def on_quotesource_editing_started(self, *unused_a):
        self.prevent_fetch()
        self.suspend_escape_handling()

    def on_quotesource_editing_done(self, cell, path, new_text, *unused_a, **unused_kw):
        thedict = dict(x for x in cell.props.model)
        try:
            new_quotesource = thedict[new_text]
        except KeyError:
            return
        self.gatherer.database.update_quoter(path, new_quotesource)
        currency = self.gatherer.database.get_currency_by_path(path)
        self.preferences["quotesources"][str(currency)] = new_text
        self.allow_fetch()
        self.resume_escape_handling()

    def on_quotesource_editing_canceled(self, *unused_a):
        self.allow_fetch()
        self.resume_escape_handling()

    def on_price_in_editing_started(self, unused_cell, entry, *unused_a):
        self.prevent_fetch()
        self.suspend_escape_handling()
        text = entry.get_text()
        text = text.split("\n")
        entry.set_text(", ".join(text))

    def on_price_in_editing_done(self, unused_cell, path, new_text, *unused_a, **unused_kw):
        new_currencies = [
            self.journal.commodity(x.strip(), True)
            for x in new_text.split(",")
        ]
        self.gatherer.database.update_price_in(path, new_currencies)
        currency = self.gatherer.database.get_currency_by_path(path)
        self.preferences["quotecurrencies"][str(currency)] = [
            str(s) for s in new_currencies
        ]
        self.allow_fetch()
        self.resume_escape_handling()

    def on_price_in_editing_canceled(self, *unused_a):
        self.allow_fetch()
        self.resume_escape_handling()

    def do_fetch(self, *unused_a):
        self.gatherer.gather_quotes()

    def save_fetched_prices(self, *unused_a):
        UpdatePricesCommon.save_fetched_prices(self)
        self.emit("delete-event", None)

    def save_preferences(self):
        self.preferences.persist()

    def run(self):
        self.connect("delete-event", Gtk.main_quit)
        GObject.idle_add(self.show_all)
        Gtk.main()


def main(argv):
    for datum in "PUBLIC_API_URL OAUTH_API_URL DATATABLES_URL".split():
        if getattr(yahoo_finance.yql, datum).startswith("http:"):
            setattr(yahoo_finance.yql, datum, "https" + yahoo_finance.yql.PUBLIC_API_URL[4:])

    p = get_argparser()
    args = p.parse_args(argv[1:])
    ledgerhelpers.enable_debugging(args.debug)

    GObject.threads_init()

    journal, settings = gui.load_journal_and_settings_for_gui(
        price_file_mandatory=True
    )
    klass = UpdatePricesApp if not args.batch else UpdatePricesCommon
    app = klass(journal, settings)
    return app.run()



from ledgerhelpers.legacy import get_terminal_width, \
print_line_ellipsized, read_one_character, go_cursor_up, blank_line, \
generate_record as generate_record_novalidate

import ledger


class LedgerParseError(ValueError):
    pass


def prompt_for_amount(fdin, fdout, prompt, commodity_example):
    cols = get_terminal_width(fdin)
    line = prompt + ("" if not commodity_example else " '': %s" % commodity_example)
    print_line_ellipsized(fdout, cols, line)
    x = []
    match = commodity_example
    while True:
        char = read_one_character(fdin)
        if char in "\n\r\t":
            break
        elif char == "\x7f":
            if x: x.pop()
        else:
            x.append(char)
        inp = "".join(x)
        try:
            match = ledger.Amount(inp) * commodity_example
        except ArithmeticError:
            try:
                match = ledger.Amount(inp)
            except ArithmeticError:
                match = ""
        cols = get_terminal_width(fdin)
        go_cursor_up(fdout)
        blank_line(fdout, cols)
        go_cursor_up(fdout)
        line = prompt + " " + "'%s': %s" % (inp, match)
        print_line_ellipsized(fdout, cols, line)
    assert match is not None
    return match


def generate_record(title, date, auxdate, state, accountamounts,
                    validate=False):
    """Generates a transaction record.
    See callee.  Validate uses Ledger to validate the record.
    """
    lines = generate_record_novalidate(
        title, date, auxdate,
        state, accountamounts
    )

    if validate:
        sess = ledger.Session()
        try:
            sess.read_journal_from_string("\n".join(lines))
        except RuntimeError as e:
            lines = [x.strip() for x in str(e).splitlines() if x.strip()]
            lines = [x for x in lines if not x.startswith("While")]
            lines = [x + ("." if not x.endswith(":") else "") for x in lines]
            lines = " ".join(lines)
            if lines:
                raise LedgerParseError(lines)
            else:
                raise LedgerParseError("Ledger could not validate this transaction")

    return lines



#!/usr/bin/python3
# coding: utf-8

from gi.repository import GObject
import gi
gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Gdk
from gi.repository import Gtk

import ledgerhelpers as h
import ledgerhelpers.legacy_needsledger as hln
from ledgerhelpers import gui
from ledgerhelpers.dateentry import DateEntry
from ledgerhelpers.transactionstatebutton import TransactionStateButton


class EditableTransactionView(Gtk.Grid):

    __gsignals__ = {
        'changed': (GObject.SIGNAL_RUN_LAST, None, ()),
        'payee-focus-out-event': (GObject.SIGNAL_RUN_LAST, None, ()),
        'payee-changed': (GObject.SIGNAL_RUN_LAST, None, ()),
    }

    css = b"""
editabletransactionview {
  border: 1px @borders inset;
  background: #fff;
}

editabletransactionview grid {
  border: none;
}

editabletransactionview entry {
  background: transparent;
  border: none;
}

editabletransactionview button {
  background: transparent;
  border: 1px solid transparent;
}
"""

    def __init__(self):
        gui.add_css(self.css)
        Gtk.Grid.__init__(self)
        self.set_column_spacing(0)

        self._postings_modified = False

        row = 0

        container = Gtk.Grid()
        container.set_hexpand(False)
        container.set_column_spacing(0)

        self.when = DateEntry()
        self.when.set_activates_default(True)
        self.when.set_hexpand(False)
        container.attach(self.when, 0, 0, 1, 1)
        self.when.connect("changed", self.child_changed)

        self.clearing_when = DateEntry()
        self.clearing_when.set_activates_default(True)
        self.clearing_when.set_hexpand(False)
        self.clearing_when.follow(self.when)
        container.attach(self.clearing_when, 1, 0, 1, 1)
        self.clearing_when.connect("changed", self.child_changed)

        self.clearing = TransactionStateButton()
        container.attach(self.clearing, 2, 0, 1, 1)
        self.clearing.connect("clicked", self.child_changed)

        self.payee = gui.EagerCompletingEntry()
        self.payee.set_hexpand(True)
        self.payee.set_activates_default(True)
        self.payee.set_size_request(300, -1)
        self.payee.set_placeholder_text(
            "Payee or description (type for completion)"
        )
        container.attach(self.payee, 3, 0, 1, 1)
        self.payee.connect("changed", self.payee_changed)
        self.payee.connect("changed", self.child_changed)
        self.payee.connect("focus-out-event", self.payee_focused_out)

        container.set_focus_chain(
            [self.when, self.clearing_when, self.clearing, self.payee]
        )

        container_evbox = Gtk.EventBox()
        container_evbox.add(container)
        self.attach(container_evbox, 0, row, 1, 1)

        row += 1

        self.lines_grid = Gtk.Grid()
        self.lines_grid.set_column_spacing(0)

        lines_evbox = Gtk.EventBox()
        lines_evbox.add(self.lines_grid)
        self.attach(lines_evbox, 0, row, 1, 1)

        self.lines = []
        self.accounts_for_completion = Gtk.ListStore(GObject.TYPE_STRING)
        self.payees_for_completion = Gtk.ListStore(GObject.TYPE_STRING)
        self.add_line()

        for x in container_evbox, lines_evbox:
            x.connect("key-press-event", self.handle_keypresses)
            x.add_events(Gdk.EventMask.KEY_PRESS_MASK)

    def handle_keypresses(self, obj, ev):
        if (
            ev.state & Gdk.ModifierType.CONTROL_MASK and
            (ev.keyval in (
                Gdk.KEY_plus, Gdk.KEY_KP_Add,
                Gdk.KEY_equal, Gdk.KEY_KP_Equal,
                Gdk.KEY_minus, Gdk.KEY_KP_Subtract, Gdk.KEY_underscore,
                Gdk.KEY_Page_Up, Gdk.KEY_KP_Page_Up,
                Gdk.KEY_Page_Down, Gdk.KEY_KP_Page_Down,
            ))
        ):
            if ev.state & Gdk.ModifierType.SHIFT_MASK:
                return self.clearing_when._on_entry__key_press_event(obj, ev)
            else:
                return self.when._on_entry__key_press_event(obj, ev)

        if (ev.state & Gdk.ModifierType.MOD1_MASK):
            keybobjects = {
                Gdk.KEY_t: self.when,
                Gdk.KEY_l: self.clearing_when,
                Gdk.KEY_p: self.payee,
            }
            for keyval, obj in list(keybobjects.items()):
                if ev.keyval == keyval:
                    obj.grab_focus()
                    return True

    def set_transaction_date(self, date):
        self.when.set_date(date)

    def set_accounts_for_completion(self, account_list):
        accounts = Gtk.ListStore(GObject.TYPE_STRING)
        [accounts.append((str(a),)) for a in account_list]
        for account, unused_amount in self.lines:
            account.get_completion().set_model(accounts)
        self.accounts_for_completion = accounts

    def set_payees_for_completion(self, payees_list):
        payees = Gtk.ListStore(GObject.TYPE_STRING)
        [payees.append((a,)) for a in payees_list]
        self.payee.get_completion().set_model(payees)
        self.payees_for_completion = payees

    def handle_data_changes(self, widget, unused_eventfocus):
        numlines = len(self.lines)
        for n, (account, amount) in reversed(list(enumerate(self.lines))):
            if n + 1 == numlines:
                continue
            p = amount.get_amount_and_price_formatted()
            if not account.get_text().strip() and not p:
                self.remove_line(n)
        last_account = self.lines[-1][0]
        last_amount = self.lines[-1][1]
        a, p = last_amount.get_amount_and_price()
        if (a or p) and last_account.get_text().strip():
            self.add_line()
        acctswidgets = dict((w[0], n) for n, w in enumerate(self.lines))
        if widget in acctswidgets:
            # If the account in the account widget has an associated
            # default commodity, then we set the default commodity of the
            # amount widget to the commodity for the account.
            account = widget.get_text().strip()
            amountwidget = self.lines[acctswidgets[widget]][1]
            c = self._get_default_commodity(account)
            if c:
                amountwidget.set_default_commodity(c)
        amtwidgets = dict((w[1], n) for n, w in enumerate(self.lines))
        if widget in amtwidgets:
            # If the amount widget has an amount, and the account associated
            # with it does not have a default commodity, and the widget itself
            # has a default commodity that differs from its current amount's
            # commodity, then we set the default commodity of the amount widget
            # to match the commodity of the amount entered in it.
            # This is extremely useful when clients of this data entry grid
            # are autofilling this grid, but haven't yet given us a default
            # commodity getter to discover account commodities.
            if widget.get_amount():
                currdef = widget.get_default_commodity()
                currcom = widget.get_amount().commodity
                accountwidget = self.lines[amtwidgets[widget]][0]
                account = accountwidget.get_text().strip()
                c = self._get_default_commodity(account)
                if not c and str(currdef) != str(currcom):
                    widget.set_default_commodity(currcom)
        if widget in [x[0] for x in self.lines] + [x[1] for x in self.lines]:
            self._postings_modified = True

    def set_default_commodity_getter(self, getter):
        """Records the new commodity getter.

        A getter is a callable that takes one account name and returns
        one commodity to be used as default for that account.  If the
        getter cannot find a default commodity, it must return None.
        """
        self._default_commodity_getter = getter
        for line in self.lines:
            accountwidget = line[0]
            amountwidget = line[1]
            account = accountwidget.get_text().strip()
            c = self._get_default_commodity(account)
            if c:
                amountwidget.set_default_commodity(c)

    def _get_default_commodity(self, account_name):
        getter = getattr(self, "_default_commodity_getter", None)
        if getter:
            return getter(account_name)

    def child_changed(self, w, unused=None):
        self.handle_data_changes(w, None)
        self.emit("changed")

    def payee_changed(self, unused_w, unused=None):
        self.emit("payee-changed")

    def payee_focused_out(self, unused_w, unused=None):
        self.emit("payee-focus-out-event")

    def get_payee_text(self):
        return self.payee.get_text()

    def remove_line(self, number):
        account, amount = self.lines[number]
        account_is_focus = account.is_focus()
        amount_is_focus = amount.is_focus()
        for hid in account._handler_ids:
            account.disconnect(hid)
        for hid in amount._handler_ids:
            amount.disconnect(hid)
        self.lines.pop(number)
        self.lines_grid.remove_row(number)
        try:
            account, amount = self.lines[number]
        except IndexError:
            account, amount = self.lines[number - 1]
        if account_is_focus:
            account.grab_focus()
        if amount_is_focus:
            amount.grab_focus()

    def postings_modified(self):
        return self._postings_modified

    def postings_empty(self):
        return (
            len(self.lines) < 2 and
            not self.lines[0][0].get_text() and
            not self.lines[0][0].get_text()
        )

    def _clear_postings(self):
        while len(self.lines) > 1:
            self.remove_line(0)
        self.lines[0][0].set_text("")
        self.lines[0][1].set_amount_and_price(None, None)

    def clear(self):
        self._clear_postings()
        self.payee.set_text("")
        self._postings_modified = False

    def set_clearing(self, clearingstate):
        self.clearing.set_state(clearingstate)

    def replace_postings(self, transactionpostings):
        """Replace postings with a list of TransactionPosting."""
        self._clear_postings()
        for n, tp in enumerate(transactionpostings):
            self.add_line()
            self.lines[n][0].set_text(tp.account)
            self.lines[n][1].set_text(tp.amount)
        self._postings_modified = False

    def add_line(self):
        account = gui.EagerCompletingEntry()
        account.set_hexpand(True)
        account.set_width_chars(40)
        account.set_activates_default(True)
        account.get_completion().set_model(self.accounts_for_completion)
        account.set_placeholder_text(
            "Account (type for completion)"
        )
        hid3 = account.connect("changed", self.child_changed)
        account._handler_ids = [hid3]

        amount = gui.LedgerAmountWithPriceEntry(display=False)
        amount.set_activates_default(True)
        amount.set_placeholder_text("Amount")
        hid3 = amount.connect("changed", self.child_changed)
        amount._handler_ids = [hid3]

        row = len(self.lines)

        if amount.display:
            amount.remove(amount.display)
            self.lines_grid.attach(amount.display, 0, row, 1, 1)
            amount.remove(amount.entry)
            self.lines_grid.attach(amount.entry, 1, row, 1, 1)
        else:
            amount.remove(amount.entry)
            self.lines_grid.attach(amount.entry, 0, row, 1, 1)
        self.lines_grid.attach(account, 2, row, 1, 1)

        account.show()
        amount.show()

        self.lines.append((account, amount))

    def title_grab_focus(self):
        self.payee.grab_focus()

    def lines_grab_focus(self):
        for account, amount in self.lines:
            if not account.get_text().strip():
                account.grab_focus()
                return
            if not amount.get_amount_and_price_formatted():
                amount.grab_focus()
                return
        else:
            if self.lines:
                self.lines[0][1].grab_focus()
            pass

    def get_data_for_transaction_record(self):
        title = self.payee.get_text().strip()
        date = self.when.get_date()
        clearing_state = self.clearing.get_state_char()
        clearing_when = self.clearing_when.get_date()

        def get_entries():
            entries = []
            for account, amount in self.lines:
                account = account.get_text().strip()
                p = amount.get_amount_and_price_formatted()
                if account or p:
                    entries.append((account, p))
            return entries

        accountamounts = [(x, y) for x, y in get_entries()]
        return title, date, clearing_when, clearing_state, accountamounts

    def validate(self, grab_focus=False):
        """Raises ValidationError if the transaction is not valid."""
        title, date, auxdate, statechar, lines = (
            self.get_data_for_transaction_record()
        )
        if not title:
            if grab_focus:
                self.payee.grab_focus()
            raise h.TransactionInputValidationError(
                "Transaction title cannot be empty"
            )
        if len(lines) < 2:
            if grab_focus:
                self.lines_grab_focus()
            raise h.TransactionInputValidationError(
                "Enter at least two transaction entries"
            )
        try:
            hln.generate_record(title, date, auxdate, statechar, lines,
                              validate=True)
        except hln.LedgerParseError as e:
            raise h.TransactionInputValidationError(str(e))


EditableTransactionView.set_css_name("editabletransactionview")



#!/usr/bin/python3

import datetime
import os
import re
import sys
sys.path.append(os.path.dirname(__file__))
import ledgerhelpers.legacy as common
from ledgerhelpers import gui


date_re = "^([0-9][0-9][0-9][0-9].[0-9][0-9].[0-9][0-9])(=[0-9][0-9][0-9][0-9].[0-9][0-9].[0-9][0-9]|)\\s+(.+)"
date_re = re.compile(date_re, re.RegexFlag.DOTALL)


def clear(f):
    changed = False
    lines = open(f).readlines()

    for n, line in enumerate(lines):
        m = date_re.match(line)
        if not m:
            continue
        if m.group(3).strip().startswith("*"):
            continue
        lines_to_write = [line]
        originaln = n
        while True:
            n = n + 1
            try:
                nextline = lines[n]
            except IndexError:
                break
            if nextline.startswith(" ") or nextline.startswith("\t"):
                lines_to_write.append(nextline)
            else:
                break
        initial_unparsed = m.group(2)[1:] if m.group(2) else m.group(1)
        initial = common.parse_date(initial_unparsed)
        if initial > datetime.date.today():
            continue
        for line in lines_to_write:
            sys.stdout.write(line)
        sys.stdout.flush()
        choice = common.prompt_for_date_optional(
            sys.stdin, sys.stdout,
            "Mark cleared at this date?",
            initial,
        )
        if choice is not None:
            choice_formatted = common.format_date(choice, initial_unparsed)
            if m.group(1) == choice_formatted:
                lines[originaln] = "%s * %s" % (
                    m.group(1),
                    m.group(3)
                )
            else:
                lines[originaln] = "%s=%s * %s" % (
                    m.group(1),
                    choice_formatted,
                    m.group(3)
                )
            for number in range(originaln + 1, n):
                # remove cleared bits on legs of the transaction
                lines[number] = re.sub("^(\\s+)\\*\\s+", "\\1", lines[number])
            changed = True
        else:
            pass
    if changed:
        y = open(f + ".new",  "w")
        y.write("".join(lines))
        y.flush()
        try:
            os.rename(f + ".new", f)
        except Exception:
            os.unlink(f + ".new")
            raise


def main():
    ledger_file = gui.find_ledger_file_for_gui()
    return clear(ledger_file)



#!/usr/bin/python3

import glob
from setuptools import setup
import os
import platform
import sys

assert sys.version_info.major == 3, "This program can no longer be built for Python 2"

dir = os.path.dirname(__file__)
path_to_main_file = os.path.join(dir, "src/ledgerhelpers/__init__.py")
path_to_readme = os.path.join(dir, "README.md")
for line in open(path_to_main_file):
	if line.startswith('__version__'):
		version = line.split()[-1].strip("'").strip('"')
		break
else:
	raise ValueError('"__version__" not found in "src/ledgerhelpers/__init__.py"')
readme = open(path_to_readme).read(-1)

classifiers = [
'Development Status :: 3 - Alpha',
'Environment :: X11 Applications :: GTK',
'Intended Audience :: End Users/Desktop',
'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
'Operating System :: POSIX :: Linux',
'Programming Language :: Python :: 3 :: Only',
'Programming Language :: Python :: 3.6',
'Topic :: Office/Business :: Financial :: Accounting',
]

programs = [
    "withdraw-cli",
    "cleartrans-cli",
    "sorttrans-cli",
    "updateprices",
    "sellstock-cli",
    "addtrans",
]

# https://github.com/Rudd-O/ledgerhelpers/issues/3
# Don't write to /usr/share/applications on OS X to work around the
# 'System Integrity Protection'.
data_files = [
	("/usr/share/applications", ["applications/%s.desktop" % p for p in programs]),
	("/usr/share/doc/ledgerhelpers/doc", glob.glob(os.path.join(os.path.dirname(__file__), "doc", "*"))),
] if platform.system() != 'Darwin' else []

setup(
	name='ledgerhelpers',
	version=version,
	description='A collection of helper programs and a helper library for Ledger (ledger-cli)',
	long_description = readme,
	long_description_content_type = "text/markdown",
	author='Manuel Amador (Rudd-O)',
	author_email='rudd-o@rudd-o.com',
	license="GPLv2+",
	url='http://github.com/Rudd-O/ledgerhelpers',
	package_dir=dict([
                    ("ledgerhelpers", "src/ledgerhelpers"),
					]),
	classifiers = classifiers,
	packages=["ledgerhelpers",
              "ledgerhelpers.programs"],
	data_files = data_files,
	scripts=["bin/%s" % p for p in programs],
	keywords="accounting ledger ledger-cli",
	requires=["ledger", "yahoo_finance"],
	zip_safe=False,
)



import datetime
import fcntl
import struct
import termios
import tty


CURSOR_UP = "\033[F"
QUIT = "quit"


yes_chars = "yY\n\r"
no_chars = "nN\x7f"
yesno_choices = dict()
for char in yes_chars:
    yesno_choices[char] = True
for char in no_chars:
    yesno_choices[char] = False
del char
del yes_chars
del no_chars


class Escaped(KeyboardInterrupt): pass


def go_cursor_up(fd):
    fd.write(CURSOR_UP)


def blank_line(fd, chars):
    fd.write(" " * chars)
    print()


def read_one_character(from_):        
    old_settings = termios.tcgetattr(from_.fileno())
    try:
        tty.setraw(from_.fileno())
        char = from_.read(1)
        if char == "\x1b":
            raise Escaped()
        if char == "\x03":
            raise KeyboardInterrupt()
    finally:
        termios.tcsetattr(from_.fileno(), termios.TCSADRAIN, old_settings)
    return char


def print_line_ellipsized(fileobj, maxlen, text):
    if len(text) > maxlen:
        text = text[:maxlen]
    fileobj.write(text)
    fileobj.flush()
    print()


def get_terminal_size(fd):
    def ioctl_GWINSZ(fd):
        return struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
    return ioctl_GWINSZ(fd)


def get_terminal_width(fd):
    return get_terminal_size(fd)[1]


def prompt_for_account(fdin, fdout, accounts, prompt, default):
    cols = get_terminal_width(fdin)
    line = prompt + ("" if not default else " '': %s" % default)
    print_line_ellipsized(fdout, cols, line)
    x = []
    match = default
    while True:
        char = read_one_character(fdin)
        if char in "\n\r\t":
            break
        elif char == "\x7f":
            if x: x.pop()
        else:
            x.append(char)
        inp = "".join(x)
        if not inp:
            match = default
        else:
            matches = [ a for a in accounts if inp.lower() in a.lower() ]
            match = matches[0] if matches else inp if inp else default
        cols = get_terminal_width(fdin)
        go_cursor_up(fdout)
        blank_line(fdout, cols)
        go_cursor_up(fdout)
        line = prompt + " " + "'%s': %s" % (inp, match)
        print_line_ellipsized(fdout, cols, line)
    return match


def choose(fdin, fdout, prompt, map_choices):
    """Based on single-char input, return a value from map_choices."""
    cols = get_terminal_width(fdin)
    line = prompt
    print_line_ellipsized(fdout, cols, line)
    while True:
        char = read_one_character(fdin)
        if char in map_choices:
            return map_choices[char]


def yesno(fdin, fdout, prompt):
    """Return True upon yY or ENTER, return False upon nN or BACKSPACE."""
    return choose(fdin, fdout, prompt, yesno_choices)


def prompt_for_expense(prompt):
    return input(prompt + " ").strip()


def prompt_for_date(fdin, fdout, prompt, initial, optional=False):
    """Return None if bool(optional) evaluates to True."""
    cols = get_terminal_width(fdin)
    if optional:
        opt = "[+/- changes, n skips, ENTER/tab accepts]"
    else:
        opt = "[+/- changes, ENTER/tab accepts]"
    line = prompt + ("" if not initial else " %s" % initial)
    print_line_ellipsized(fdout, cols, line + " " + opt)
    while True:
        char = read_one_character(fdin)
        if char in "\n\r\t":
            break
        elif char == "+":
            initial = initial + datetime.timedelta(1)
        elif char == "-":
            initial = initial + datetime.timedelta(-1)
        elif char in "nN" and optional:
            return None
        cols = get_terminal_width(fdin)
        go_cursor_up(fdout)
        blank_line(fdout, cols)
        go_cursor_up(fdout)
        line = prompt + " " + "%s" % initial
        print_line_ellipsized(fdout, cols, line + " " + opt)
    return initial


def prompt_for_date_optional(fdin, fdout, prompt, initial):
    return prompt_for_date(fdin, fdout, prompt, initial, True)


def parse_date(putative_date, return_format=False):
    """Returns a date substring in a ledger entry, parsed as datetime.date."""
    # FIXME: use Ledger functions to parse dates, not mine.
    formats = ["%Y-%m-%d", "%Y/%m/%d"]
    for f in formats:
        try:
            d = datetime.datetime.strptime(putative_date, f).date()
            break
        except ValueError as e:
            last_exception = e
            continue
    try:
        if return_format:
            return d, f
        else:
            return d
    except UnboundLocalError:
        raise ValueError("cannot parse date from format %s: %s" % (f, last_exception))


def format_date(date_obj, sample_date):
    _, fmt = parse_date(sample_date, True)
    return date_obj.strftime(fmt)


def generate_record(title, date, auxdate, state, accountamounts):
    """Generates a transaction record.

    date is a datetime.date
    title is a string describing the title of the transaction
    auxdate is the date when the transaction cleared, or None
    statechar is a char from parser.CHAR_* or empty string
    accountamounts is a list of:
    (account, amount)
    """
    def stramt(amt):
        assert type(amt) not in (tuple, list), amt
        if not amt:
            return ""
        return str(amt).strip()

    if state:
        state = state + " "
    else:
        state = ""

    lines = [""]
    linesemptyamts = []
    if auxdate:
        if auxdate != date:
            lines.append("%s=%s %s%s" % (date, auxdate, state, title))
        else:
            lines.append("%s %s%s" % (date, state, title))
    else:
        lines.append("%s %s%s" % (date, state, title))

    try:
        longest_acct = max(list(len(a) for a, _ in accountamounts))
        longest_amt = max(list(len(stramt(am)) for _, am in accountamounts))
    except ValueError:
        longest_acct = 30
        longest_amt = 30
    pattern = "    %-" + str(longest_acct) + "s    %" + str(longest_amt) + "s"
    pattern2 = "    %-" + str(longest_acct) + "s"
    for account, amount in accountamounts:
        if stramt(amount):
            lines.append(pattern % (account, stramt(amount)))
        else:
            linesemptyamts.append((pattern2 % (account,)).rstrip())
    lines = lines + linesemptyamts
    lines.append("")
    return lines



#!/usr/bin/python3

import datetime
import ledger
import os
import sys
import ledgerhelpers
import ledgerhelpers.legacy as common
import ledgerhelpers.legacy_needsledger as common2
import ledgerhelpers.journal as journal


def main():
    s = ledgerhelpers.Settings.load_or_defaults(os.path.expanduser("~/.ledgerhelpers.ini"))
    j = journal.Journal.from_file(ledgerhelpers.find_ledger_file(), None)
    accts, commodities = j.accounts_and_last_commodity_for_account()

    when = common.prompt_for_date(
        sys.stdin, sys.stdout,
        "When?",
        s.get("last_date", datetime.date.today())
    )
    if when == datetime.date.today():
        del s["last_date"]
    else:
        s["last_date"] = when

    asset1 = common.prompt_for_account(
        sys.stdin, sys.stdout,
        accts, "From where?",
        s.get("last_withdrawal_account", None)
    )
    assert asset1, "Not an account: %s" % asset1
    s["last_withdrawal_account"] = asset1
    asset1_currency = commodities.get(asset1, ledger.Amount("$ 1"))

    asset2 = common.prompt_for_account(
        sys.stdin, sys.stdout,
        accts, "To where?",
        s.get("last_deposit_account", None)
    )
    assert asset2, "Not an account: %s" % asset2
    s["last_deposit_account"] = asset2
    asset2_currency = commodities.get(asset2, ledger.Amount("$ 1"))

    amount1 = common2.prompt_for_amount(
        sys.stdin, sys.stdout,
        "How much?", asset1_currency
    )

    amount2 = common2.prompt_for_amount(
        sys.stdin, sys.stdout,
        "What was deposited?", asset2_currency
    )

    lines = j.generate_record("Withdrawal", when, None, "", [
        (asset1, -1 * amount1),
        (asset2, amount2),
    ])
    print("========== Record ==========")
    print("\n".join(lines))
    save = common.yesno(
        sys.stdin, sys.stderr,
        "Hit ENTER or y to save it to the file, BACKSPACE or n to skip saving: "
    )
    if save:
        j.add_text_to_file(lines)
